"""
@file
@brief Modified converter from
`LightGbm.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
lightgbm/operator_converters/LightGbm.py>`_.
"""
from collections import Counter
import copy
import numbers
import pprint
import numpy
from onnx import TensorProto
from skl2onnx.common._apply_operation import apply_div, apply_reshape, apply_sub  # pylint: disable=E0611
from skl2onnx.common.tree_ensemble import get_default_tree_classifier_attribute_pairs
from skl2onnx.proto import onnx_proto
from skl2onnx.common.shape_calculator import (
    calculate_linear_regressor_output_shapes,
    calculate_linear_classifier_output_shapes)
from skl2onnx.common.data_types import guess_numpy_type
from skl2onnx.common.tree_ensemble import sklearn_threshold
from ..sklconv.tree_converters import _fix_tree_ensemble
from ..helpers.lgbm_helper import (
    dump_lgbm_booster, modify_tree_for_rule_in_set)


def calculate_lightgbm_output_shapes(operator):
    """
    Shape calculator for LightGBM Booster
    (see :epkg:`lightgbm`).
    """
    op = operator.raw_operator
    if hasattr(op, "_model_dict"):
        objective = op._model_dict['objective']
    elif hasattr(op, 'objective_'):
        objective = op.objective_
    else:
        raise RuntimeError(  # pragma: no cover
            "Unable to find attributes '_model_dict' or 'objective_' in "
            "instance of type %r (list of attributes=%r)." % (
                type(op), dir(op)))
    if objective.startswith('binary') or objective.startswith('multiclass'):
        return calculate_linear_classifier_output_shapes(operator)
    if objective.startswith('regression'):  # pragma: no cover
        return calculate_linear_regressor_output_shapes(operator)
    raise NotImplementedError(  # pragma: no cover
        f"Objective '{objective}' is not implemented yet.")


def _translate_split_criterion(criterion):
    # If the criterion is true, LightGBM use the left child. Otherwise, right child is selected.
    if criterion == '<=':
        return 'BRANCH_LEQ'
    if criterion == '<':  # pragma: no cover
        return 'BRANCH_LT'
    if criterion == '>=':  # pragma: no cover
        return 'BRANCH_GTE'
    if criterion == '>':  # pragma: no cover
        return 'BRANCH_GT'
    if criterion == '==':  # pragma: no cover
        return 'BRANCH_EQ'
    if criterion == '!=':  # pragma: no cover
        return 'BRANCH_NEQ'
    raise ValueError(  # pragma: no cover
        'Unsupported splitting criterion: %s. Only <=, '
        '<, >=, and > are allowed.')


def _create_node_id(node_id_pool):
    i = 0
    while i in node_id_pool:
        i += 1
    node_id_pool.add(i)
    return i


def _parse_tree_structure(tree_id, class_id, learning_rate,
                          tree_structure, attrs):
    """
    The pool of all nodes' indexes created when parsing a single tree.
    Different tree use different pools.
    """
    node_id_pool = set()
    node_pyid_pool = dict()

    node_id = _create_node_id(node_id_pool)
    node_pyid_pool[id(tree_structure)] = node_id

    # The root node is a leaf node.
    if ('left_child' not in tree_structure or
            'right_child' not in tree_structure):
        _parse_node(tree_id, class_id, node_id, node_id_pool, node_pyid_pool,
                    learning_rate, tree_structure, attrs)
        return

    left_pyid = id(tree_structure['left_child'])
    right_pyid = id(tree_structure['right_child'])

    if left_pyid in node_pyid_pool:
        left_id = node_pyid_pool[left_pyid]
        left_parse = False
    else:
        left_id = _create_node_id(node_id_pool)
        node_pyid_pool[left_pyid] = left_id
        left_parse = True

    if right_pyid in node_pyid_pool:
        right_id = node_pyid_pool[right_pyid]
        right_parse = False
    else:
        right_id = _create_node_id(node_id_pool)
        node_pyid_pool[right_pyid] = right_id
        right_parse = True

    attrs['nodes_treeids'].append(tree_id)
    attrs['nodes_nodeids'].append(node_id)

    attrs['nodes_featureids'].append(tree_structure['split_feature'])
    mode = _translate_split_criterion(tree_structure['decision_type'])
    attrs['nodes_modes'].append(mode)

    if isinstance(tree_structure['threshold'], str):
        try:  # pragma: no cover
            th = float(tree_structure['threshold'])  # pragma: no cover
        except ValueError as e:  # pragma: no cover
            text = pprint.pformat(tree_structure)
            if len(text) > 99999:
                text = text[:99999] + "\n..."
            raise TypeError("threshold must be a number not '{}'"
                            "\n{}".format(tree_structure['threshold'], text)) from e
    else:
        th = tree_structure['threshold']
    if mode == 'BRANCH_LEQ':
        th2 = sklearn_threshold(th, numpy.float32, mode)
    else:
        # other decision criteria are not implemented
        th2 = th
    attrs['nodes_values'].append(th2)

    # Assume left is the true branch and right is the false branch
    attrs['nodes_truenodeids'].append(left_id)
    attrs['nodes_falsenodeids'].append(right_id)
    if tree_structure['default_left']:
        # attrs['nodes_missing_value_tracks_true'].append(1)
        if (tree_structure["missing_type"] in ('None', None) and
                float(tree_structure['threshold']) < 0.0):
            attrs['nodes_missing_value_tracks_true'].append(0)
        else:
            attrs['nodes_missing_value_tracks_true'].append(1)
    else:
        attrs['nodes_missing_value_tracks_true'].append(0)
    attrs['nodes_hitrates'].append(1.)
    if left_parse:
        _parse_node(
            tree_id, class_id, left_id, node_id_pool, node_pyid_pool,
            learning_rate, tree_structure['left_child'], attrs)
    if right_parse:
        _parse_node(
            tree_id, class_id, right_id, node_id_pool, node_pyid_pool,
            learning_rate, tree_structure['right_child'], attrs)


def _parse_node(tree_id, class_id, node_id, node_id_pool, node_pyid_pool,
                learning_rate, node, attrs):
    """
    Parses nodes.
    """
    if ((hasattr(node, 'left_child') and hasattr(node, 'right_child')) or
            ('left_child' in node and 'right_child' in node)):

        left_pyid = id(node['left_child'])
        right_pyid = id(node['right_child'])

        if left_pyid in node_pyid_pool:
            left_id = node_pyid_pool[left_pyid]
            left_parse = False
        else:
            left_id = _create_node_id(node_id_pool)
            node_pyid_pool[left_pyid] = left_id
            left_parse = True

        if right_pyid in node_pyid_pool:
            right_id = node_pyid_pool[right_pyid]
            right_parse = False
        else:
            right_id = _create_node_id(node_id_pool)
            node_pyid_pool[right_pyid] = right_id
            right_parse = True

        attrs['nodes_treeids'].append(tree_id)
        attrs['nodes_nodeids'].append(node_id)

        attrs['nodes_featureids'].append(node['split_feature'])
        attrs['nodes_modes'].append(
            _translate_split_criterion(node['decision_type']))
        if isinstance(node['threshold'], str):
            try:  # pragma: no cover
                attrs['nodes_values'].append(  # pragma: no cover
                    float(node['threshold']))
            except ValueError as e:  # pragma: no cover
                text = pprint.pformat(node)
                if len(text) > 99999:
                    text = text[:99999] + "\n..."
                raise TypeError("threshold must be a number not '{}'"
                                "\n{}".format(node['threshold'], text)) from e
        else:
            attrs['nodes_values'].append(node['threshold'])

        # Assume left is the true branch and right is the false branch
        attrs['nodes_truenodeids'].append(left_id)
        attrs['nodes_falsenodeids'].append(right_id)
        if node['default_left']:
            # attrs['nodes_missing_value_tracks_true'].append(1)
            if (node['missing_type'] in ('None', None) and
                    float(node['threshold']) < 0.0):
                attrs['nodes_missing_value_tracks_true'].append(0)
            else:
                attrs['nodes_missing_value_tracks_true'].append(1)
        else:
            attrs['nodes_missing_value_tracks_true'].append(0)
        attrs['nodes_hitrates'].append(1.)

        # Recursively dive into the child nodes
        if left_parse:
            _parse_node(
                tree_id, class_id, left_id, node_id_pool, node_pyid_pool,
                learning_rate, node['left_child'], attrs)
        if right_parse:
            _parse_node(
                tree_id, class_id, right_id, node_id_pool, node_pyid_pool,
                learning_rate, node['right_child'], attrs)
    elif hasattr(node, 'left_child') or hasattr(node, 'right_child'):
        raise ValueError('Need two branches')  # pragma: no cover
    else:
        # Node attributes
        attrs['nodes_treeids'].append(tree_id)
        attrs['nodes_nodeids'].append(node_id)
        attrs['nodes_featureids'].append(0)
        attrs['nodes_modes'].append('LEAF')
        # Leaf node has no threshold. A zero is appended but it will never be used.
        attrs['nodes_values'].append(0.)
        # Leaf node has no child. A zero is appended but it will never be used.
        attrs['nodes_truenodeids'].append(0)
        # Leaf node has no child. A zero is appended but it will never be used.
        attrs['nodes_falsenodeids'].append(0)
        # Leaf node has no split function. A zero is appended but it will never be used.
        attrs['nodes_missing_value_tracks_true'].append(0)
        attrs['nodes_hitrates'].append(1.)

        # Leaf attributes
        attrs['class_treeids'].append(tree_id)
        attrs['class_nodeids'].append(node_id)
        attrs['class_ids'].append(class_id)
        attrs['class_weights'].append(
            float(node['leaf_value']) * learning_rate)


def _split_tree_ensemble_atts(attrs, split):
    """
    Splits the attributes of a TreeEnsembleRegressor into
    multiple trees in order to do the summation in double instead of floats.
    """
    trees_id = list(sorted(set(attrs['nodes_treeids'])))
    results = []
    index = 0
    while index < len(trees_id):
        index2 = min(index + split, len(trees_id))
        subset = set(trees_id[index: index2])

        indices_node = []
        indices_target = []
        for j, v in enumerate(attrs['nodes_treeids']):
            if v in subset:
                indices_node.append(j)
        for j, v in enumerate(attrs['target_treeids']):
            if v in subset:
                indices_target.append(j)

        if (len(indices_node) >= len(attrs['nodes_treeids']) or
                len(indices_target) >= len(attrs['target_treeids'])):
            raise RuntimeError(  # pragma: no cover
                "Initial attributes are not consistant."
                "\nindex=%r index2=%r subset=%r"
                "\nnodes_treeids=%r\ntarget_treeids=%r"
                "\nindices_node=%r\nindices_target=%r" % (
                    index, index2, subset,
                    attrs['nodes_treeids'], attrs['target_treeids'],
                    indices_node, indices_target))

        ats = {}
        for name, att in attrs.items():
            if name == 'nodes_treeids':
                new_att = [att[i] for i in indices_node]
                new_att = [i - att[0] for i in new_att]
            elif name == 'target_treeids':
                new_att = [att[i] for i in indices_target]
                new_att = [i - att[0] for i in new_att]
            elif name.startswith("nodes_"):
                new_att = [att[i] for i in indices_node]
                assert len(new_att) == len(indices_node)
            elif name.startswith("target_"):
                new_att = [att[i] for i in indices_target]
                assert len(new_att) == len(indices_target)
            elif name == 'name':
                new_att = f"{att}{len(results)}"
            else:
                new_att = att
            ats[name] = new_att

        results.append(ats)
        index = index2

    return results


def convert_lightgbm(scope, operator, container):  # pylint: disable=R0914
    """
    This converters reuses the code from
    `LightGbm.py <https://github.com/onnx/onnxmltools/blob/master/onnxmltools/convert/
    lightgbm/operator_converters/LightGbm.py>`_ and makes
    some modifications. It implements converters
    for models in :epkg:`lightgbm`.
    """
    verbose = getattr(container, 'verbose', 0)
    gbm_model = operator.raw_operator
    if hasattr(gbm_model, '_model_dict_info'):
        gbm_text, info = gbm_model._model_dict_info
    else:
        if verbose >= 2:
            print("[convert_lightgbm] dump_model")  # pragma: no cover
        gbm_text, info = dump_lgbm_booster(gbm_model.booster_, verbose=verbose)
    opsetml = container.target_opset_all.get('ai.onnx.ml', None)
    if opsetml is None:
        opsetml = 3 if container.target_opset >= 16 else 1
    if verbose >= 2:
        print(  # pragma: no cover
            "[convert_lightgbm] modify_tree_for_rule_in_set")
    modify_tree_for_rule_in_set(gbm_text, use_float=True, verbose=verbose,
                                info=info)

    attrs = get_default_tree_classifier_attribute_pairs()
    attrs['name'] = operator.full_name

    # Create different attributes for classifier and
    # regressor, respectively
    post_transform = None
    if gbm_text['objective'].startswith('binary'):
        n_classes = 1
        attrs['post_transform'] = 'LOGISTIC'
    elif gbm_text['objective'].startswith('multiclass'):
        n_classes = gbm_text['num_class']
        attrs['post_transform'] = 'SOFTMAX'
    elif gbm_text['objective'].startswith('regression'):
        n_classes = 1  # Regressor has only one output variable
        attrs['post_transform'] = 'NONE'
        attrs['n_targets'] = n_classes
    elif gbm_text['objective'].startswith(('poisson', 'gamma')):
        n_classes = 1  # Regressor has only one output variable
        attrs['n_targets'] = n_classes
        # 'Exp' is not a supported post_transform value in the ONNX spec yet,
        # so we need to add an 'Exp' post transform node to the model
        attrs['post_transform'] = 'NONE'
        post_transform = "Exp"
    else:
        raise RuntimeError(  # pragma: no cover
            "LightGBM objective should be cleaned already not '{}'.".format(
                gbm_text['objective']))

    # Use the same algorithm to parse the tree
    if verbose >= 2:  # pragma: no cover
        from tqdm import tqdm
        loop = tqdm(gbm_text['tree_info'])
        loop.set_description("parse")
    else:
        loop = gbm_text['tree_info']
    for i, tree in enumerate(loop):
        tree_id = i
        class_id = tree_id % n_classes
        # tree['shrinkage'] --> LightGbm provides figures with it already.
        learning_rate = 1.
        _parse_tree_structure(
            tree_id, class_id, learning_rate, tree['tree_structure'], attrs)

    if verbose >= 2:
        print("[convert_lightgbm] onnx")  # pragma: no cover
    # Sort nodes_* attributes. For one tree, its node indexes
    # should appear in an ascent order in nodes_nodeids. Nodes
    # from a tree with a smaller tree index should appear
    # before trees with larger indexes in nodes_nodeids.
    node_numbers_per_tree = Counter(attrs['nodes_treeids'])
    tree_number = len(node_numbers_per_tree.keys())
    accumulated_node_numbers = [0] * tree_number
    for i in range(1, tree_number):
        accumulated_node_numbers[i] = (
            accumulated_node_numbers[i - 1] + node_numbers_per_tree[i - 1])
    global_node_indexes = []
    for i in range(len(attrs['nodes_nodeids'])):
        tree_id = attrs['nodes_treeids'][i]
        node_id = attrs['nodes_nodeids'][i]
        global_node_indexes.append(
            accumulated_node_numbers[tree_id] + node_id)
    for k, v in attrs.items():
        if k.startswith('nodes_'):
            merged_indexes = zip(
                copy.deepcopy(global_node_indexes), v)
            sorted_list = [pair[1]
                           for pair in sorted(merged_indexes,
                                              key=lambda x: x[0])]
            attrs[k] = sorted_list

    dtype = guess_numpy_type(operator.inputs[0].type)
    if dtype != numpy.float64:
        dtype = numpy.float32

    if dtype == numpy.float64:
        for key in ['nodes_values', 'nodes_hitrates', 'target_weights',
                    'class_weights', 'base_values']:
            if key not in attrs:
                continue
            attrs[key] = numpy.array(attrs[key], dtype=dtype)

    # Create ONNX object
    if (gbm_text['objective'].startswith('binary') or
            gbm_text['objective'].startswith('multiclass')):
        # Prepare label information for both of TreeEnsembleClassifier
        # and ZipMap
        class_type = onnx_proto.TensorProto.STRING  # pylint: disable=E1101
        zipmap_attrs = {'name': scope.get_unique_variable_name('ZipMap')}
        if all(isinstance(i, (numbers.Real, bool, numpy.bool_))
               for i in gbm_model.classes_):
            class_type = onnx_proto.TensorProto.INT64  # pylint: disable=E1101
            class_labels = [int(i) for i in gbm_model.classes_]
            attrs['classlabels_int64s'] = class_labels
            zipmap_attrs['classlabels_int64s'] = class_labels
        elif all(isinstance(i, str) for i in gbm_model.classes_):
            class_labels = [str(i) for i in gbm_model.classes_]
            attrs['classlabels_strings'] = class_labels
            zipmap_attrs['classlabels_strings'] = class_labels
        else:
            raise ValueError(  # pragma: no cover
                'Only string and integer class labels are allowed')

        # Create tree classifier
        probability_tensor_name = scope.get_unique_variable_name(
            'probability_tensor')
        label_tensor_name = scope.get_unique_variable_name('label_tensor')

        if dtype == numpy.float64 and opsetml < 3:
            container.add_node('TreeEnsembleClassifierDouble', operator.input_full_names,
                               [label_tensor_name, probability_tensor_name],
                               op_domain='mlprodict', op_version=1, **attrs)
        else:
            container.add_node('TreeEnsembleClassifier', operator.input_full_names,
                               [label_tensor_name, probability_tensor_name],
                               op_domain='ai.onnx.ml', op_version=1, **attrs)

        prob_tensor = probability_tensor_name

        if gbm_model.boosting_type == 'rf':
            col_index_name = scope.get_unique_variable_name('col_index')
            first_col_name = scope.get_unique_variable_name('first_col')
            zeroth_col_name = scope.get_unique_variable_name('zeroth_col')
            denominator_name = scope.get_unique_variable_name('denominator')
            modified_first_col_name = scope.get_unique_variable_name(
                'modified_first_col')
            unit_float_tensor_name = scope.get_unique_variable_name(
                'unit_float_tensor')
            merged_prob_name = scope.get_unique_variable_name('merged_prob')
            predicted_label_name = scope.get_unique_variable_name(
                'predicted_label')
            classes_name = scope.get_unique_variable_name('classes')
            final_label_name = scope.get_unique_variable_name('final_label')

            container.add_initializer(
                col_index_name, onnx_proto.TensorProto.INT64, [], [1])  # pylint: disable=E1101
            container.add_initializer(
                unit_float_tensor_name, onnx_proto.TensorProto.FLOAT, [], [1.0])  # pylint: disable=E1101
            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT, [], [100.0])  # pylint: disable=E1101
            container.add_initializer(classes_name, class_type,
                                      [len(class_labels)], class_labels)

            container.add_node(
                'ArrayFeatureExtractor',
                [probability_tensor_name, col_index_name],
                first_col_name,
                name=scope.get_unique_operator_name(
                    'ArrayFeatureExtractor'),
                op_domain='ai.onnx.ml')
            apply_div(scope, [first_col_name, denominator_name],
                      modified_first_col_name, container, broadcast=1)
            apply_sub(
                scope, [unit_float_tensor_name, modified_first_col_name],
                zeroth_col_name, container, broadcast=1)
            container.add_node(
                'Concat', [zeroth_col_name, modified_first_col_name],
                merged_prob_name,
                name=scope.get_unique_operator_name('Concat'), axis=1)
            container.add_node(
                'ArgMax', merged_prob_name,
                predicted_label_name,
                name=scope.get_unique_operator_name('ArgMax'), axis=1)
            container.add_node(
                'ArrayFeatureExtractor', [classes_name, predicted_label_name],
                final_label_name,
                name=scope.get_unique_operator_name('ArrayFeatureExtractor'),
                op_domain='ai.onnx.ml')
            apply_reshape(scope, final_label_name,
                          operator.outputs[0].full_name,
                          container, desired_shape=[-1, ])
            prob_tensor = merged_prob_name
        else:
            container.add_node('Identity', label_tensor_name,
                               operator.outputs[0].full_name,
                               name=scope.get_unique_operator_name('Identity'))

        # Convert probability tensor to probability map
        # (keys are labels while values are the associated probabilities)
        container.add_node('Identity', prob_tensor,
                           operator.outputs[1].full_name)
    else:
        # Create tree regressor
        output_name = scope.get_unique_variable_name('output')

        keys_to_be_renamed = list(
            k for k in attrs if k.startswith('class_'))

        for k in keys_to_be_renamed:
            # Rename class_* attribute to target_*
            # because TreeEnsebmleClassifier
            # and TreeEnsembleClassifier have different ONNX attributes
            attrs['target' + k[5:]] = copy.deepcopy(attrs[k])
            del attrs[k]

        options = container.get_options(gbm_model, dict(split=-1))
        split = options['split']
        if split == -1:
            if dtype == numpy.float64 and opsetml < 3:
                container.add_node(
                    'TreeEnsembleRegressorDouble', operator.input_full_names,
                    output_name, op_domain='mlprodict', op_version=1, **attrs)
            else:
                container.add_node(
                    'TreeEnsembleRegressor', operator.input_full_names,
                    output_name, op_domain='ai.onnx.ml', op_version=1, **attrs)
        else:
            tree_attrs = _split_tree_ensemble_atts(attrs, split)
            tree_nodes = []
            for i, ats in enumerate(tree_attrs):
                tree_name = scope.get_unique_variable_name('tree%d' % i)
                if dtype == numpy.float64:
                    container.add_node(
                        'TreeEnsembleRegressorDouble', operator.input_full_names,
                        tree_name, op_domain='mlprodict', op_version=1, **ats)
                    tree_nodes.append(tree_name)
                else:
                    container.add_node(
                        'TreeEnsembleRegressor', operator.input_full_names,
                        tree_name, op_domain='ai.onnx.ml', op_version=1, **ats)
                    cast_name = scope.get_unique_variable_name('dtree%d' % i)
                    container.add_node(
                        'Cast', tree_name, cast_name, to=TensorProto.DOUBLE,  # pylint: disable=E1101
                        name=scope.get_unique_operator_name("dtree%d" % i))
                    tree_nodes.append(cast_name)
            if dtype == numpy.float64:
                container.add_node(
                    'Sum', tree_nodes, output_name,
                    name=scope.get_unique_operator_name(f"sumtree{len(tree_nodes)}"))
            else:
                cast_name = scope.get_unique_variable_name('ftrees')
                container.add_node(
                    'Sum', tree_nodes, cast_name,
                    name=scope.get_unique_operator_name(f"sumtree{len(tree_nodes)}"))
                container.add_node(
                    'Cast', cast_name, output_name, to=TensorProto.FLOAT,  # pylint: disable=E1101
                    name=scope.get_unique_operator_name("dtree%d" % i))

        if gbm_model.boosting_type == 'rf':
            denominator_name = scope.get_unique_variable_name('denominator')

            container.add_initializer(
                denominator_name, onnx_proto.TensorProto.FLOAT,  # pylint: disable=E1101
                [], [100.0])

            apply_div(scope, [output_name, denominator_name],
                      operator.output_full_names, container, broadcast=1)
        elif post_transform:
            container.add_node(
                post_transform, output_name,
                operator.output_full_names,
                name=scope.get_unique_operator_name(
                    post_transform))
        else:
            container.add_node('Identity', output_name,
                               operator.output_full_names,
                               name=scope.get_unique_operator_name('Identity'))
    if opsetml >= 3:
        _fix_tree_ensemble(scope, container, opsetml, dtype)
    if verbose >= 2:
        print("[convert_lightgbm] end")  # pragma: no cover



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *onnx_conv*.
"""
import warnings
import numbers
import numpy
from skl2onnx import update_registered_converter
from skl2onnx.common.shape_calculator import (
    calculate_linear_classifier_output_shapes,
    calculate_linear_regressor_output_shapes)
from .scorers import register_scorers


def _custom_parser_xgboost(scope, model, inputs, custom_parsers=None):
    """
    Custom parser for *XGBClassifier* and *LGBMClassifier*.
    """
    if custom_parsers is not None and model in custom_parsers:
        return custom_parsers[model](
            scope, model, inputs, custom_parsers=custom_parsers)
    if not all(isinstance(i, (numbers.Real, bool, numpy.bool_))
               for i in model.classes_):
        raise NotImplementedError(  # pragma: no cover
            "Current converter does not support string labels.")
    try:
        from skl2onnx._parse import _parse_sklearn_classifier
    except ImportError as e:  # pragma: no cover
        import skl2onnx
        raise ImportError(
            "Hidden API has changed in module 'skl2onnx=={}', "
            "installation path is '{}'.".format(
                skl2onnx.__version__, skl2onnx.__file__)) from e
    return _parse_sklearn_classifier(scope, model, inputs)


def _register_converters_lightgbm(exc=True):
    """
    This functions registers additional converters
    for :epkg:`lightgbm`.

    @param      exc     if True, raises an exception if a converter cannot
                        registered (missing package for example)
    @return             list of models supported by the new converters
    """
    registered = []

    try:
        from lightgbm import LGBMClassifier
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register LGBMClassifier due to '{e}'.")
            LGBMClassifier = None
    if LGBMClassifier is not None:
        try:
            from skl2onnx._parse import _parse_sklearn_classifier
        except ImportError as e:  # pragma: no cover
            import skl2onnx
            raise ImportError(
                "Hidden API has changed in module 'skl2onnx=={}', "
                "installation path is '{}'.".format(
                    skl2onnx.__version__, skl2onnx.__file__)) from e
        from .operator_converters.conv_lightgbm import (
            convert_lightgbm, calculate_lightgbm_output_shapes)
        update_registered_converter(
            LGBMClassifier, 'LgbmClassifier',
            calculate_lightgbm_output_shapes,
            convert_lightgbm, parser=_parse_sklearn_classifier,
            options={'zipmap': [True, False], 'nocl': [True, False]})
        registered.append(LGBMClassifier)

    try:
        from lightgbm import LGBMRegressor
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register LGBMRegressor due to '{e}'.")
            LGBMRegressor = None
    if LGBMRegressor is not None:
        from .operator_converters.conv_lightgbm import convert_lightgbm
        update_registered_converter(
            LGBMRegressor, 'LightGbmLGBMRegressor',
            calculate_linear_regressor_output_shapes,
            convert_lightgbm, options={'split': None})
        registered.append(LGBMRegressor)

    try:
        from lightgbm import Booster
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register LGBMRegressor due to '{e}'.")
            Booster = None
    if Booster is not None:
        from .operator_converters.conv_lightgbm import (
            convert_lightgbm, calculate_lightgbm_output_shapes)
        from .operator_converters.parse_lightgbm import (
            lightgbm_parser, WrappedLightGbmBooster,
            WrappedLightGbmBoosterClassifier,
            shape_calculator_lightgbm_concat,
            converter_lightgbm_concat,
            MockWrappedLightGbmBoosterClassifier)
        update_registered_converter(
            Booster, 'LightGbmBooster', calculate_lightgbm_output_shapes,
            convert_lightgbm, parser=lightgbm_parser,
            options={'cast': [True, False]})
        update_registered_converter(
            WrappedLightGbmBooster, 'WrapperLightGbmBooster',
            calculate_lightgbm_output_shapes,
            convert_lightgbm, parser=lightgbm_parser)
        update_registered_converter(
            WrappedLightGbmBoosterClassifier, 'WrappedLightGbmBoosterClassifier',
            calculate_lightgbm_output_shapes,
            convert_lightgbm, parser=lightgbm_parser,
            options={'zipmap': [True, False], 'nocl': [True, False]})
        update_registered_converter(
            MockWrappedLightGbmBoosterClassifier, 'MockWrappedLightGbmBoosterClassifier',
            calculate_lightgbm_output_shapes,
            convert_lightgbm, parser=lightgbm_parser)
        update_registered_converter(
            None, 'LightGBMConcat',
            shape_calculator_lightgbm_concat,
            converter_lightgbm_concat)
        registered.append(Booster)
        registered.append(WrappedLightGbmBooster)
        registered.append(WrappedLightGbmBoosterClassifier)

    return registered


def _register_converters_xgboost(exc=True):
    """
    This functions registers additional converters
    for :epkg:`xgboost`.

    @param      exc     if True, raises an exception if a converter cannot
                        registered (missing package for example)
    @return             list of models supported by the new converters
    """
    registered = []

    try:
        from xgboost import XGBClassifier
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register XGBClassifier due to '{e}'.")
            XGBClassifier = None
    if XGBClassifier is not None:
        from .operator_converters.conv_xgboost import convert_xgboost
        update_registered_converter(
            XGBClassifier, 'XGBoostXGBClassifier',
            calculate_linear_classifier_output_shapes,
            convert_xgboost, parser=_custom_parser_xgboost,
            options={'zipmap': [True, False], 'raw_scores': [True, False],
                     'nocl': [True, False]})
        registered.append(XGBClassifier)

    try:
        from xgboost import XGBRegressor
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register LGBMRegressor due to '{e}'.")
            XGBRegressor = None
    if XGBRegressor is not None:
        from .operator_converters.conv_xgboost import convert_xgboost
        update_registered_converter(XGBRegressor, 'XGBoostXGBRegressor',
                                    calculate_linear_regressor_output_shapes,
                                    convert_xgboost)
        registered.append(XGBRegressor)
    return registered


def _register_converters_mlinsights(exc=True):
    """
    This functions registers additional converters
    for :epkg:`mlinsights`.

    @param      exc     if True, raises an exception if a converter cannot
                        registered (missing package for example)
    @return             list of models supported by the new converters
    """
    registered = []

    try:
        from mlinsights.mlmodel import TransferTransformer
    except ImportError as e:  # pragma: no cover
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register models from 'mlinsights' due to '{e}'.")
            TransferTransformer = None

    if TransferTransformer is not None:
        from .operator_converters.conv_transfer_transformer import (
            shape_calculator_transfer_transformer, convert_transfer_transformer,
            parser_transfer_transformer)
        update_registered_converter(
            TransferTransformer, 'MlInsightsTransferTransformer',
            shape_calculator_transfer_transformer,
            convert_transfer_transformer,
            parser=parser_transfer_transformer,
            options='passthrough')
        registered.append(TransferTransformer)

    return registered


def _register_converters_skl2onnx(exc=True):
    """
    This functions registers additional converters
    for :epkg:`skl2onnx`.

    @param      exc     if True, raises an exception if a converter cannot
                        registered (missing package for example)
    @return             list of models supported by the new converters
    """
    registered = []

    try:
        import skl2onnx.sklapi.register  # pylint: disable=W0611
        from skl2onnx.sklapi import WOETransformer
        model = [WOETransformer]
    except ImportError as e:  # pragma: no cover
        try:
            import skl2onnx
            from pyquickhelper.texthelper.version_helper import (
                compare_module_version)
            if compare_module_version(skl2onnx.__version__, '1.9.3') < 0:
                # Too old version of skl2onnx.
                return []
        except ImportError:
            pass
        if exc:
            raise e
        else:
            warnings.warn(
                f"Cannot register models from 'skl2onnx' due to {e!r}.")
            model = None

    if model is not None:
        registered.extend(model)
    return registered


def register_converters(exc=True):
    """
    This functions registers additional converters
    to the list of converters :epkg:`sklearn-onnx` declares.

    @param      exc     if True, raises an exception if a converter cannot
                        registered (missing package for example)
    @return             list of models supported by the new converters
    """
    ext = _register_converters_lightgbm(exc=exc)
    ext += _register_converters_xgboost(exc=exc)
    ext += _register_converters_mlinsights(exc=exc)
    ext += _register_converters_skl2onnx(exc=exc)
    ext += register_scorers()
    return ext



"""
@file
@brief Functions to simplify, compress an ONNX graph.

.. versionadded:: 0.9
"""
import logging
from onnx import ModelProto, GraphProto, FunctionProto
from onnx.helper import (
    make_function, make_model, make_value_info, make_graph,
    make_tensor_type_proto, make_node, make_operatorsetid)


logger = logging.getLogger('onnx:compress')


def _check_expression(expe):
    att = expe.attribute[0].g
    inputs = [i.name for i in att.input]
    if list(expe.input) != inputs:
        raise RuntimeError(  # pragma: no cover
            f'Name mismatch in node Expression {expe.input!r} != {inputs!r}.')
    outputs = [o.name for o in att.output]
    if list(expe.output) != outputs:
        raise RuntimeError(  # pragma: no cover
            f'Name mismatch in node Expression {expe.input!r} != {inputs!r}.')


def _fuse_node(o, node, node_next):
    """
    Merges two nodes having one input/output in common.

    :param o: output name
    :param node: first node (it outputs the results)
    :param node_next: second node (it ingests the result)
    :return: merged node
    """
    type_expression = ('mlprodict', 'Expression')
    if list(node.output) != [o]:
        raise RuntimeError(  # pragma: no cover
            f"The only output of the first node should be {[o]!r} not {node.output!r}.")
    cannot_do = {('', 'If'), ('', 'Loop'), ('', 'Scan')}
    key1 = node.domain, node.op_type
    if key1 in cannot_do:
        return None
    key2 = node_next.domain, node_next.op_type
    if key2 in cannot_do:
        return None

    if key1 == type_expression:
        _check_expression(node)
    if key2 == type_expression:
        _check_expression(node_next)

    graph = None

    if node.domain == '' and node_next.domain == '':
        # Simple case
        inputs = [make_value_info(name, make_tensor_type_proto(0, []))
                  for name in node.input]
        outputs = [make_value_info(name, make_tensor_type_proto(0, []))
                   for name in node_next.output]
        graph = make_graph([node, node_next], "expression", inputs, outputs)

    elif key1 == type_expression and node_next.domain == '':
        att = node.attribute[0].g
        inputs = att.input
        outputs = [make_value_info(name, make_tensor_type_proto(0, []))
                   for name in node_next.output]
        graph = make_graph(list(att.node) + [node_next],
                           "expression", inputs, outputs)

    elif node.domain == '' and key2 == type_expression:
        att = node_next.attribute[0].g
        inputs = [make_value_info(name, make_tensor_type_proto(0, []))
                  for name in node.input]
        outputs = att.output
        graph = make_graph([node] + list(att.node),
                           "expression", inputs, outputs)

    elif key1 == type_expression and key2 == type_expression:
        att1 = node.attribute[0].g
        att2 = node_next.attribute[0].g
        inputs = att1.input
        outputs = att2.output
        graph = make_graph(list(att1.node) + list(att2.node),
                           "expression", inputs, outputs)

    if graph is not None:
        new_node = make_node(
            'Expression', node.input, node_next.output, domain='mlprodict',
            expression=graph)
        return new_node

    raise NotImplementedError(  # pragma: no cover
        "Unable to merge nodes '%s/%s' and '%s/%s'." % (
            node.domain, node.op_type, node_next.domain, node_next.op_type))


def _compress_nodes_once(nodes, verbose=0):
    """
    Compresses a sequence of node to make it more
    readable. If possible, it creates a node `Expression`
    with a graph as an attribute.

    :param nodes: sequence of nodes to compress
    :return: compressed sequence of nodes
    """
    # check that a result is used only once
    order = {}
    results = {}
    for node in nodes:
        order[id(node)] = (len(order), node)
        for name in node.input:
            if name in results:
                results[name] += 1
            else:
                results[name] = 1

    once = {k: v for k, v in results.items() if v == 1}
    if len(once) == 0:
        return nodes

    once_nodes_o = {}
    once_nodes_i = {}
    for node in nodes:
        if len(node.output) != 1:
            continue
        for o in node.output:
            if o in once:
                once_nodes_o[o] = node
        for i in node.input:
            if i in once:
                once_nodes_i[i] = node

    if len(once_nodes_o) == 0:
        return nodes

    if verbose > 0:
        logger.debug(
            "Results to compress: %r", list(sorted(once_nodes_o)))

    while len(once_nodes_o) > 0:
        o, node = once_nodes_o.popitem()
        node_next = once_nodes_i[o]
        new_node = _fuse_node(o, node, node_next)
        if new_node is None:
            # nothing can be done
            continue
        once_nodes_o.update({o: new_node for o in node_next.output
                             if o in once_nodes_o})
        once_nodes_i.update({i: new_node for i in node.input
                             if i in once_nodes_i})
        order[id(new_node)] = (order[id(node)][0], new_node)
        del order[id(node)]
        del order[id(node_next)]

    ordered = list(sorted((v[0], k, v[1]) for k, v in order.items()))
    return [v[-1] for v in ordered]


def _compress_nodes(nodes, verbose=0):
    """
    Compresses a sequence of node to make it more
    readable. If possible, it creates a node `Expression`
    with a graph as an attribute.

    :param nodes: sequence of nodes to compress
    :return: compressed sequence of nodes
    """
    return _compress_nodes_once(nodes, verbose=verbose)


def compress_proto(proto, verbose=0):
    """
    Compresses a :epkg:`ModelProto`, :epkg:`FunctionProto`,
    :epkg:`GraphProto`. The function detects nodes outputting
    results only used once. It then fuses it with the node
    taking it as an input.

    :param proto: :epkg:`ModelProto`, :epkg:`FunctionProto`,
        :epkg:`GraphProto`
    :param verbose: logging
    :return: same type

    .. versionadded:: 0.9
    """
    if isinstance(proto, FunctionProto):
        nodes = _compress_nodes(proto.node, verbose=verbose)
        if len(nodes) == len(proto.node):
            # unchanged
            return proto
        if verbose:
            logger.debug(
                "Compressed function %r/%r from %d nodes to %d.",
                proto.domain, proto.name, len(proto.node), len(nodes))
        opsets = {op.domain: op.version for op in proto.opset_import}
        opsets['mlprodict'] = 1

        return make_function(
            proto.domain, proto.name,
            proto.input, proto.output, nodes,
            opset_imports=[
                make_operatorsetid(k, v) for k, v in opsets.items()],
            attributes=proto.attribute,
            doc_string=proto.doc_string)

    if isinstance(proto, ModelProto):
        modified = 0
        new_graph = compress_proto(proto.graph, verbose=verbose)
        if id(new_graph) != id(proto.graph):
            modified += 1
        fcts = []
        for f in proto.functions:
            new_f = compress_proto(f, verbose=verbose)
            if id(new_f) != id(f):
                modified += 1
            fcts.append(new_f)
        if modified == 0:
            return proto
        opsets = {op.domain: op.version for op in proto.opset_import}
        opsets['mlprodict'] = 1
        if verbose:
            logger.debug(
                "Compressed model %s modified=%d.", proto.name, modified)
        return make_model(
            new_graph, functions=fcts,
            opset_imports=[
                make_operatorsetid(k, v) for k, v in opsets.items()],
            producer_name=proto.producer_name,
            producer_version=proto.producer_version,
            ir_version=proto.ir_version,
            doc_string=proto.doc_string,
            domain=proto.domain,
            model_version=proto.model_version)

    if isinstance(proto, GraphProto):
        nodes = _compress_nodes(proto.node, verbose=verbose)
        if len(nodes) == len(proto.node):
            # unchanged
            return proto
        if verbose:
            logger.debug(
                "Compressed graph %s from %d nodes to %d.",
                proto.name, len(proto.node), len(nodes))
        return make_graph(
            nodes, proto.name, proto.input, proto.output,
            proto.initializer, sparse_initializer=proto.sparse_initializer)

    raise TypeError(  # pragma: no cover
        "Unexpected type for proto %r, it should ModelProto, "
        "GraphProto or FunctionProto." % type(proto))



# -*- encoding: utf-8 -*-
"""
@file
@brief Shortcut to *ops_cpu*.
"""
import textwrap
from ..excs import MissingOperatorError
from ._op import OpRunCustom
from ._op_list import __dict__ as d_op_list


_additional_ops = {}


def register_operator(cls, name=None, overwrite=True):
    """
    Registers a new runtime operator.

    @param      cls         class
    @param      name        by default ``cls.__name__``,
                            or *name* if defined
    @param      overwrite   overwrite or raise an exception
    """
    if name is None:
        name = cls.__name__
    if name not in _additional_ops:
        _additional_ops[name] = cls
    elif not overwrite:
        raise RuntimeError(  # pragma: no cover
            "Unable to overwrite existing operator '{}': {} "
            "by {}".format(name, _additional_ops[name], cls))


def load_op(onnx_node, desc=None, options=None, runtime=None):
    """
    Gets the operator related to the *onnx* node.

    :param onnx_node: :epkg:`onnx` node
    :param desc: internal representation
    :param options: runtime options
    :param runtime: runtime
    :param existing_functions: existing functions
    :return: runtime class
    """
    from ... import __max_supported_opset__
    if desc is None:
        raise ValueError("desc should not be None.")  # pragma no cover
    name = onnx_node.op_type
    opset = options.get('target_opset', None) if options is not None else None
    current_opset = __max_supported_opset__
    chosen_opset = current_opset
    if opset == current_opset:
        opset = None
    if opset is not None:
        if not isinstance(opset, int):
            raise TypeError(  # pragma no cover
                f"opset must be an integer not {type(opset)}")
        name_opset = name + "_" + str(opset)
        for op in range(opset, 0, -1):
            nop = name + "_" + str(op)
            if nop in d_op_list:
                name_opset = nop
                chosen_opset = op
                break
    else:
        name_opset = name

    if name_opset in _additional_ops:
        cl = _additional_ops[name_opset]
    elif name in _additional_ops:
        cl = _additional_ops[name]
    elif name_opset in d_op_list:
        cl = d_op_list[name_opset]
    elif name in d_op_list:
        cl = d_op_list[name]
    else:
        raise MissingOperatorError(  # pragma no cover
            "Operator '{}' from domain '{}' has no runtime yet. "
            "Available list:\n"
            "{} - {}".format(
                name, onnx_node.domain,
                "\n".join(sorted(_additional_ops)),
                "\n".join(textwrap.wrap(
                    " ".join(
                        _ for _ in sorted(d_op_list)
                        if "_" not in _ and _ not in {'cl', 'clo', 'name'})))))

    if hasattr(cl, 'version_higher_than'):
        opv = min(current_opset, chosen_opset)
        if cl.version_higher_than > opv:
            # The chosen implementation does not support
            # the opset version, we need to downgrade it.
            if ('target_opset' in options and
                    options['target_opset'] is not None):  # pragma: no cover
                raise RuntimeError(
                    "Supported version {} > {} (opset={}) required version, "
                    "unable to find an implementation version {} found "
                    "'{}'\n--ONNX--\n{}\n--AVAILABLE--\n{}".format(
                        cl.version_higher_than, opv, opset,
                        options['target_opset'], cl.__name__, onnx_node,
                        "\n".join(
                            _ for _ in sorted(d_op_list)
                            if "_" not in _ and _ not in {'cl', 'clo', 'name'})))
            options = options.copy()
            options['target_opset'] = current_opset
            return load_op(onnx_node, desc=desc, options=options)

    if options is None:
        options = {}  # pragma: no cover
    return cl(onnx_node, desc=desc, runtime=runtime, **options)



"""
@file
@brief Xop API to build onnx graphs. Inspired from :epkg:`sklearn-onnx`.

.. versionadded:: 0.9
"""
import numpy
from onnx import ValueInfoProto
from onnx.helper import make_tensor_type_proto
from onnx.mapping import NP_TYPE_TO_TENSOR_TYPE
from onnx.defs import onnx_opset_version
from .. import __max_supported_opset__


def max_supported_opset():
    """
    Returns the latest supported opset for the main domain.

    .. runpython::
        :showcode:

        from mlprodict.npy.xop_variable import max_supported_opset
        print("max_supported_opset() returns", max_supported_opset())
    """
    return min(__max_supported_opset__, onnx_opset_version())


def is_numpy_dtype(dtype):
    """
    Tells if a dtype is a numpy dtype.

    :param dtype: anything
    :return: boolean
    """
    if isinstance(dtype, (list, dict, Variable)):
        return False
    if dtype in NP_TYPE_TO_TENSOR_TYPE:
        return True
    dt = numpy.dtype(dtype)
    if dt in NP_TYPE_TO_TENSOR_TYPE:
        return True
    return False


def numpy_type_prototype(dtype):
    """
    Converts a numpy dtyp into a TensorProto dtype.

    :param dtype: dtype
    :return: proto dtype
    """
    if dtype in NP_TYPE_TO_TENSOR_TYPE:
        return NP_TYPE_TO_TENSOR_TYPE[dtype]
    dt = numpy.dtype(dtype)
    if dt in NP_TYPE_TO_TENSOR_TYPE:
        return NP_TYPE_TO_TENSOR_TYPE[dt]
    raise ValueError(  # pragma: no cover
        f"Unable to convert dtype {dtype!r} into ProtoType.")


def guess_numpy_type(data_type):
    """
    Guesses the corresponding numpy type based on data_type.
    """
    if data_type in (numpy.float64, numpy.float32, numpy.int8, numpy.uint8,
                     numpy.str_, numpy.bool_, numpy.int32, numpy.int64):
        return data_type
    if data_type == str:
        return numpy.str_
    if data_type == bool:
        return numpy.bool_
    name2numpy = {
        'FloatTensorType': numpy.float32,
        'DoubleTensorType': numpy.float64,
        'Int32TensorType': numpy.int32,
        'Int64TensorType': numpy.int64,
        'StringTensorType': numpy.str_,
        'BooleanTensorType': numpy.bool_,
        'Complex64TensorType': numpy.complex64,
        'Complex128TensorType': numpy.complex128,
    }
    cl_name = data_type.__class__.__name__
    if cl_name in name2numpy:
        return name2numpy[cl_name]
    if hasattr(data_type, 'type'):
        return guess_numpy_type(data_type.type)
    raise NotImplementedError(  # pragma: no cover
        f"Unsupported data_type '{data_type}'.")


class ExistingVariable:
    """
    Temporary name.

    :param name: variable name
    :param op: operator it comes from
    """

    def __init__(self, name, op):
        self.name = name
        self.op = op

    def __repr__(self):
        "usual"
        return f"{self.__class__.__name__}({self.name!r})"

    @property
    def dtype(self):
        "Unknown type, returns None."
        return None

    @property
    def added_dtype(self):
        "Unknown type, returns None."
        return None


class Variable:
    """
    An input or output to an ONNX graph.

    :param name: name
    :param dtype: :epkg:`numpy` dtype (can be None)
    :param shape: shape (can be None)
    :param added_dtype: :epkg:`numpy` dtype specified at conversion type
        (can be None)
    :param added_shape: :epkg:`numpy` shape specified at conversion type
        (can be None)
    """

    def __init__(self, name, dtype=None, shape=None, added_dtype=None,
                 added_shape=None):
        if (dtype is not None and isinstance(
                dtype, (int, Variable, tuple, numpy.ndarray))):
            raise TypeError(
                f"Unexpected type {type(dtype)!r} for dtype.")
        if (added_dtype is not None and isinstance(
                added_dtype, (int, Variable, tuple, numpy.ndarray))):
            raise TypeError(
                f"Unexpected type {type(added_dtype)!r} for added_dtype.")
        if shape is not None and not isinstance(shape, (tuple, list)):
            raise TypeError(
                f"Unexpected type {type(shape)!r} for shape.")
        if (added_shape is not None and not isinstance(
                added_shape, (tuple, list))):
            raise TypeError(
                f"Unexpected type {type(added_shape)!r} for added_shape.")

        if isinstance(name, Variable):
            if (dtype is not None or shape is not None or
                    added_dtype is not None or added_shape is not None):
                raise ValueError(  # pragma: no cover
                    "If name is a Variable, then all others attributes "
                    "should be None.")

            self.name_ = name.name_
            self.dtype_ = name.dtype_
            self.added_dtype_ = name.added_dtype_
            self.shape_ = name.shape_
            self.added_shape_ = name.added_shape_
        else:
            if not isinstance(name, str):
                raise TypeError(  # pragma: no cover
                    f"name must be a string not {type(name)!r}.")

            self.name_ = name
            self.dtype_ = dtype
            self.added_dtype_ = added_dtype
            self.shape_ = shape
            self.added_shape_ = added_shape

    def to_skl2onnx(self, scope=None):
        """
        Converts this instance into an instance of *Variable*
        from :epkg:`sklearn-onnx`.
        """
        from skl2onnx.common._topology import Variable as skl2onnxVariable  # delayed
        from skl2onnx.common.data_types import _guess_numpy_type  # delayed
        inst = _guess_numpy_type(self.dtype, self.shape)
        var = skl2onnxVariable(self.name, self.name, type=inst, scope=scope)
        return var

    @staticmethod
    def from_skl2onnx(var):
        """
        Converts variable from :epkg:`sklearn-onnx` into this class.
        """
        return Variable(var.onnx_name, guess_numpy_type(var.type),
                        shape=var.type.shape)

    @staticmethod
    def from_skl2onnx_tuple(var):
        """
        Converts variable from :epkg:`sklearn-onnx` into this class
        defined as a tuple.
        """
        return Variable(var[0], guess_numpy_type(var[1]),
                        shape=var[1].shape)

    @property
    def name(self):
        "Returns the variable name (`self.name_`)."
        return self.name_

    @property
    def dtype(self):
        "Returns `self.dtype_`."
        return self.dtype_

    @property
    def added_dtype(self):
        "Returns `self.added_dtype_`."
        return self.added_dtype_

    @property
    def shape(self):
        "Returns `self.shape_`."
        return self.shape_

    @property
    def proto_type(self):
        "Returns the proto type for `self.dtype_`."
        if self.dtype_ is None:
            return 0
        return numpy_type_prototype(self.dtype_)

    @property
    def proto_added_type(self):
        "Returns the proto type for `self.added_dtype_` or `self.dtype_`."
        dt = self.added_dtype_ or self.dtype_
        if dt is None:
            return 0
        return numpy_type_prototype(dt)

    @property
    def proto_added_shape(self):
        "Returns the shape for `self.added_shape_` or `self.shape`."
        dt = self.added_shape_ or self.shape_
        if dt is None:
            return None
        return list(dt)

    def __repr__(self):
        "usual"
        kwargs = dict(dtype=self.dtype_, shape=self.shape_,
                      added_dtype=self.added_dtype_,
                      added_shape=self.added_shape_)
        kwargs = {k: v for k, v in kwargs.items() if v is not None}
        if len(kwargs) > 0:
            msg = ", " + ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
        else:
            msg = ''
        return f"{self.__class__.__name__}({self.name_!r}{msg})"

    def is_named(self, name):
        "Tells the variable is named like that."
        if not isinstance(name, str):
            raise TypeError(  # pragma: no cover
                f"name is expected to be a string not {type(name)!r}.")
        return self.name == name

    def copy_add(self, dtype):
        """
        Returns a copy of this variable with a new dtype.

        :param dtype: added type
        :return: @see cl Variable
        """
        if self.added_dtype_ is not None:
            raise RuntimeError(  # pragma: no cover
                "Cannot copy as added_dtype is not None.")
        if isinstance(dtype, numpy.ndarray):
            dtype, shape = dtype.dtype, dtype.shape
        else:
            shape = None
        return Variable(self.name_, self.dtype_, self.shape_, dtype, shape)

    def copy_merge(self, var, shape=None):
        """
        Merges information from both Variable.
        """
        if not isinstance(var, Variable):
            if shape is not None:
                raise RuntimeError(  # pragma: no cover
                    "shape must be None if var is a Variable.")
            return self.copy_add(var)
        res = Variable(self.name_, self.dtype_,
                       shape or self.shape_, self.added_dtype_,
                       self.added_shape_)
        if self.added_dtype_ is None and var.dtype_ is not None:
            res.added_dtype_ = var.dtype_
        if self.added_shape_ is None and var.shape_ is not None:
            res.added_shape_ = var.shape_
        return res

    def copy_name(self, name):
        """
        Returns a copy with a new name.
        """
        return Variable(
            name or self.name_, self.dtype_,
            self.shape_, self.added_dtype_,
            self.added_shape_)

    def __eq__(self, other):
        """
        Compares every attributes.
        """
        if not isinstance(other, Variable):
            raise TypeError(
                f"Unexpected type {type(other)!r}.")
        if self.name != other.name:
            return False
        if self.shape_ != other.shape_:
            return False
        if self.dtype_ != other.dtype_:
            return False
        return True

    def make_value_info(self):
        """
        Converts the variable into `onnx.ValueInfoProto`.

        :return: instance of `onnx.ValueInfoProto`
        """
        value_info = ValueInfoProto()
        value_info.name = self.name
        tensor_type_proto = make_tensor_type_proto(self.proto_type, self.shape)
        value_info.type.CopyFrom(tensor_type_proto)  # pylint: disable=E1101
        return value_info

    @staticmethod
    def from_pb(obj):
        """
        Creates a Variable from a protobuf object.

        :param obj: initializer, tensor
        :return: @see cl Variable
        """
        from ..onnx_tools.onnx2py_helper import from_pb
        name, ty, shape = from_pb(obj)
        return Variable(name, ty, shape=shape)


class NodeResultName:
    """
    Defines a result name for a node.

    :param node: node it comes from
    :param index: index of the output
    """

    def __init__(self, node, index):
        self.node = node
        self.index = index

    def __repr__(self):
        "Usual"
        return f"{self.__class__.__name__}({self.node!r}, {self.index!r})"

    def get_name(self):
        """
        Returns a name from output_names or a suggestion for a name.
        """
        if self.node is None:
            raise RuntimeError(  # pragma: no cover
                "node must not be None.")
        if self.node.output_names is not None:
            return self.node.output_names[self.index].name
        cl = self.node.op_type.lower()[:3]
        return "out_%s_%d" % (cl, self.index)


class DetectedVariable:
    """
    Wrapper around a @see cl Variable to detect inputs
    and outputs of a graph.

    :param node: node where the variable was detected
    :param var: instance of @see cl Variable
    :param index: index, only used if it is an output
    """

    def __init__(self, node, var, index):
        if not isinstance(var, (Variable, ExistingVariable)):
            raise TypeError(  # pragma: no cover
                f"Unexpected type {type(var)!r}, it should be a Variable.")
        self.node = node
        self.var = var
        self.index = index

    @property
    def name(self):
        "Returns variable name."
        return self.var.name

    def __repr__(self):
        "usual"
        sindex = f", {self.index}" if self.index >= 0 else ""
        if self.node is None:
            return f"{self.__class__.__name__}(None, {self.var!r}{sindex})"
        return "%s(%s-%d, %r%s)" % (
            self.__class__.__name__, self.node.__class__.__name__,
            id(self.node), self.var, sindex)


class InputDetectedVariable(DetectedVariable):
    """
    Instance of @see cl DetectedVariable.
    Only for inputs.
    """

    def __init__(self, node, var):
        DetectedVariable.__init__(self, node, var, -1)


class OutputDetectedVariable(DetectedVariable):
    """
    Instance of @see cl DetectedVariable.
    Only for outputs.
    """
    pass



"""
@file
@brief :epkg:`numpy` functions implemented with :epkg:`onnx`.

.. versionadded:: 0.6
"""
from .xop_convert import OnnxSubEstimator
from .onnx_variable import MultiOnnxVar, OnnxVar


def linear_regression(x, *, model=None):
    """
    Returns a linear regression converted into ONNX.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of :epkg:`sklearn:linear_model:LinearRegression`
    :return: instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    """
    return OnnxVar(model, x, op=OnnxSubEstimator)


def logistic_regression(x, *, model=None):
    """
    Returns a logistic regression converted into ONNX,
    option *zipmap* is set to false.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of :epkg:`sklearn:linear_model:LinearRegression`
    :return: instance of :class:`MultiOnnxVar
        <mlprodict.npy.onnx_variable.MultiOnnxVar>`, first
        output is labels, second one is the probabilities
    """
    return MultiOnnxVar(model, x, op=OnnxSubEstimator,
                        options={'zipmap': False})


def classifier(x, *, model=None):
    """
    Returns any classifier from :epkg:`scikit-learn`
    converted into ONNX assuming a converter is registered
    with :epkg:`sklearn-onnx`. Option *zipmap* is set to false.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of a classifier
    :return: instance of :class:`MultiOnnxVar
        <mlprodict.npy.onnx_variable.MultiOnnxVar>`, first
        output is labels, second one is the probabilities
    """
    return MultiOnnxVar(model, x, op=OnnxSubEstimator,
                        options={'zipmap': False})


def cluster(x, *, model=None):
    """
    Returns any cluster from :epkg:`scikit-learn`
    converted into ONNX assuming a converter is registered
    with :epkg:`sklearn-onnx`. Option *zipmap* is set to false.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of a cluster
    :return: instance of :class:`MultiOnnxVar
        <mlprodict.npy.onnx_variable.MultiOnnxVar>`, first
        output is labels, second one is the probabilities
    """
    return MultiOnnxVar(model, x, op=OnnxSubEstimator)


def regressor(x, *, model=None):
    """
    Returns any regressor from :epkg:`scikit-learn`
    converted into ONNX assuming a converter is registered
    with :epkg:`sklearn-onnx`.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of a regressor
    :return: instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    """
    return OnnxVar(model, x, op=OnnxSubEstimator)


def transformer(x, *, model=None):
    """
    Returns any transformer from :epkg:`scikit-learn`
    converted into ONNX assuming a converter is registered
    with :epkg:`sklearn-onnx`.

    :param x: array, variable name, instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    :param model: instance of a transformer
    :return: instance of :class:`OnnxVar
        <mlprodict.npy.onnx_variable.OnnxVar>`
    """
    return OnnxVar(model, x, op=OnnxSubEstimator)



"""
@file
@brief Calls all possible :epkg:`ONNX` optimisations.
"""
from .onnx_optimisation import onnx_remove_node


def onnx_optimisations(onnx_model, recursive=True, debug_info=None, **options):
    """
    Calls several possible optimisations including
    @see fn onnx_remove_node.

    @param      onnx_model      onnx model
    @param      recursive       looks into subgraphs
    @param      debug_info      debug information (private)
    @param      options         additional options
    @return                     new onnx _model
    """
    new_model = onnx_remove_node(
        onnx_model, recursive=recursive, debug_info=debug_info,
        **options)
    return new_model



# -*- coding: utf-8 -*-
import sys
import os
import platform
import warnings
from setuptools import setup, Extension, find_packages
from pyquicksetup import read_version, read_readme, default_cmdclass

#########
# settings
#########

project_var_name = "mlprodict"
versionPython = f"{sys.version_info.major}.{sys.version_info.minor}"
path = "Lib/site-packages/" + project_var_name
readme = 'README.rst'
history = "HISTORY.rst"
requirements = None

KEYWORDS = [project_var_name, 'Xavier Dupré', 'onnx', 'scikit-learn',
            'production', 'machine learning']

DESCRIPTION = ("Python Runtime for ONNX models, other helpers to convert "
               "machine learned models in C++.")
CLASSIFIERS = [
    'Programming Language :: Python :: 3',
    'Intended Audience :: Developers',
    'Intended Audience :: Science/Research',
    'License :: OSI Approved :: MIT License',
    "Operating System :: Microsoft :: Windows",
    "Operating System :: Unix",
    "Operating System :: MacOS",
    'Topic :: Software Development',
    'Topic :: Scientific/Engineering',
    'Topic :: Education',
    'Development Status :: 5 - Production/Stable',
]


#######
# data
#######

packages = find_packages()
package_dir = {k: os.path.join('.', k.replace(".", "/")) for k in packages}
package_data = {
    project_var_name + ".asv_benchmark": ["*.json"],
    project_var_name + ".npy": ["ort_get_all_operator_schema.txt"],
    project_var_name + ".onnxrt.ops_cpu": ["*.cpp", "*.hpp"],
    project_var_name + ".onnxrt.validate.data": ["*.csv"],
    project_var_name + ".onnx_tools": ["*.tmpl"],
    project_var_name + ".testing": ["*.cpp", "*.hpp"],
}


class get_pybind_include(object):
    """
    Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    `Source <https://github.com/pybind/python_example/blob/master/setup.py>`_.
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def get_compile_args():
    if sys.platform.startswith("win"):
        libraries_thread = ['kernel32']
        extra_compile_args = ['/EHsc', '/O2', '/Gy', '/openmp']
        extra_link_args = None
        define_macros = [('USE_OPENMP', None)]
    elif sys.platform.startswith("darwin"):
        libraries_thread = None
        extra_compile_args = ['-stdlib=libc++', '-mmacosx-version-min=10.7',
                              '-fpermissive', '-std=c++11',
                              '-Xpreprocessor', '-fopenmp']
        extra_link_args = ['-lomp']
        define_macros = [('USE_OPENMP', None)]
    else:
        libraries_thread = None
        # , '-o2', '-mavx512f']
        extra_compile_args = ['-fopenmp']
        extra_link_args = ['-lgomp']
        define_macros = [('USE_OPENMP', None)]
        if "Ubuntu" in platform.version() and "16.04" in platform.version():
            extra_compile_args.append('-std=c++11')
    return (libraries_thread, extra_compile_args,
            extra_link_args, define_macros)


def get_extensions():
    import numpy

    root = os.path.abspath(os.path.dirname(__file__))
    (libraries_thread, extra_compile_args,
     extra_link_args, define_macros) = get_compile_args()

    ext_roi_align = Extension(
        'mlprodict.onnxrt.ops_cpu.op_roi_align_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_roi_align_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_grid_sample = Extension(
        'mlprodict.onnxrt.ops_cpu.op_grid_sample_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_grid_sample_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_max_pool = Extension(
        'mlprodict.onnxrt.ops_cpu.op_max_pool_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_max_pool_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_gather = Extension(
        'mlprodict.onnxrt.ops_cpu.op_gather_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_gather_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_tree_ensemble_classifier = Extension(
        'mlprodict.onnxrt.ops_cpu.op_tree_ensemble_classifier_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_tree_ensemble_classifier_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_tree_ensemble_regressor = Extension(
        'mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_tree_ensemble_regressor_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_tree_ensemble_regressor_p = Extension(
        'mlprodict.onnxrt.ops_cpu.op_tree_ensemble_regressor_p_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_tree_ensemble_regressor_p_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_svm_regressor = Extension(
        'mlprodict.onnxrt.ops_cpu.op_svm_regressor_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_svm_regressor_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_tfidfvectorizer = Extension(
        'mlprodict.onnxrt.ops_cpu.op_tfidfvectorizer_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_tfidfvectorizer_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_tree_ensemble_classifier_p = Extension(
        'mlprodict.onnxrt.ops_cpu.op_tree_ensemble_classifier_p_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_tree_ensemble_classifier_p_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_svm_classifier = Extension(
        'mlprodict.onnxrt.ops_cpu.op_svm_classifier_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_svm_classifier_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    op_onnx_numpy = Extension(
        'mlprodict.onnxrt.ops_cpu._op_onnx_numpy',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/_op_onnx_numpy.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_num_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_conv = Extension(
        'mlprodict.onnxrt.ops_cpu.op_conv_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_qlinearconv = Extension(
        'mlprodict.onnxrt.ops_cpu.op_qlinear_conv_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_qlinear_conv_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(
             root, 'mlprodict/onnxrt/ops_cpu/op_qlinear_cpp_qgemm_tester_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp')
         ],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_conv_helper = Extension(
        'mlprodict.onnxrt.ops_cpu.op_conv_helper_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_helper_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_conv_transpose = Extension(
        'mlprodict.onnxrt.ops_cpu.op_conv_transpose_',
        [os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_transpose_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_conv_matrices_.cpp'),
         os.path.join(root, 'mlprodict/onnxrt/ops_cpu/op_common_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    ext_experimental_c = Extension(
        'mlprodict.testing.experimental_c_impl.experimental_c',
        [os.path.join(
            root, 'mlprodict/testing/experimental_c_impl/experimental_c.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/testing/experimental_c_impl')
        ],
        define_macros=define_macros,
        language='c++')

    ext_non_max_suppression = Extension(
        'mlprodict.onnxrt.ops_cpu.op_non_max_suppression_',
        [os.path.join(
            root, 'mlprodict/onnxrt/ops_cpu/op_non_max_suppression_.cpp')],
        extra_compile_args=extra_compile_args,
        extra_link_args=extra_link_args,
        include_dirs=[
            # Path to pybind11 headers
            get_pybind_include(),
            get_pybind_include(user=True),
            os.path.join(root, 'mlprodict/onnxrt/ops_cpu')
        ],
        define_macros=define_macros,
        language='c++')

    cython_ext = [
        Extension("mlprodict.testing.einsum.direct_blas_lapack",
                  ['mlprodict/testing/einsum/direct_blas_lapack.pyx'],
                  include_dirs=[numpy.get_include()],
                  language='c')
    ]

    from Cython.Build import cythonize
    opts = dict(boundscheck=False, cdivision=True,
                wraparound=False, language_level=3,
                cdivision_warnings=False, embedsignature=True,
                initializedcheck=False)
    cy_ext_blas = cythonize(cython_ext, compiler_directives=opts)

    ext_modules = [
        ext_conv,
        ext_conv_helper,
        ext_conv_transpose,
        ext_experimental_c,
        ext_gather,
        ext_grid_sample,
        ext_max_pool,
        ext_non_max_suppression,
        ext_qlinearconv,
        ext_roi_align,
        ext_svm_classifier,
        ext_svm_regressor,
        ext_tfidfvectorizer,
        ext_tree_ensemble_classifier,
        ext_tree_ensemble_classifier_p,
        ext_tree_ensemble_regressor,
        ext_tree_ensemble_regressor_p,
        op_onnx_numpy,
    ]
    ext_modules.extend(cy_ext_blas)
    return ext_modules


try:
    ext_modules = get_extensions()
except ImportError as e:
    warnings.warn(
        f"Unable to build C++ extension with missing dependencies {e!r}.")
    ext_modules = None

# setup

setup(
    name=project_var_name,
    ext_modules=ext_modules,
    version=read_version(__file__, project_var_name),
    author='Xavier Dupré',
    author_email='xavier.dupre@gmail.com',
    license="MIT",
    url=f"http://www.xavierdupre.fr/app/{project_var_name}/helpsphinx/index.html",
    download_url=f"https://github.com/sdpython/{project_var_name}/",
    description=DESCRIPTION,
    long_description=read_readme(__file__),
    cmdclass=default_cmdclass(),
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    packages=packages,
    package_dir=package_dir,
    package_data=package_data,
    setup_requires=["pybind11", "numpy", "onnx>=1.7.0", "scikit-learn>=0.23",
                    "jinja2", 'cython', 'pyquicksetup'],
    install_requires=["pybind11", "numpy>=1.17", "onnx>=1.7.0", 'scipy>=1.0.0',
                      'jinja2', 'cython'],
    extras_require={
        'npy': ['scikit-learn>=0.24', 'skl2onnx>=1.10.2'],
        'onnx_conv': [
            'scikit-learn>=0.24', 'skl2onnx>=1.10.2', 'lightgbm',
            'joblib', 'threadpoolctl', 'mlinsights>=0.3', 'xgboost'],
        'onnx_val': [
            'scikit-learn>=0.24', 'skl2onnx>=1.10.2',
            'onnxruntime>=1.10.0', 'joblib', 'threadpoolctl'],
        'sklapi': [
            'scikit-learn>=0.24', 'joblib', 'threadpoolctl',
            'onnxruntime>=1.19.0', 'onnxruntime-extensions'],
        'all': [
            'scikit-learn>=0.24', 'skl2onnx>=1.10.2',
            'onnxruntime>=1.10.0', 'scipy' 'joblib', 'pandas',
            'threadpoolctl', 'mlinsights>=0.3', 'lightgbm',
            'xgboost', 'mlstatpy>=0.3.593', 'onnxruntime-extensions'],
    },
)



"""
@file
@brief OnnxInferenceNode definition.
"""
import sys
import pprint
import numpy
from onnx import GraphProto, onnx_pb as onnx_proto
from onnx.onnx_cpp2py_export.defs import SchemaError  # pylint: disable=E0401,E0611
from ..onnx_tools.onnx2py_helper import get_onnx_schema
from .excs import MissingOperatorError
from .ops import load_op


class OnnxInferenceNode:
    """
    A node to execute.

    :param onnx_node: onnx_node
    :param desc: internal description
    :param global_index: it is a function which returns a unique index
        for the output this operator generates
    """
    class OnnxInferenceWrapper:
        """
        Wraps @see cl OnnxInference in a wrapper and exposes
        the necessary function.

        :param oinf: instance of @see cl OnnxInference
        """

        def __init__(self, oinf):
            if oinf is None:
                raise ValueError(  # pragma: no cover
                    "oinf cannot be None.")
            self.oinf = oinf

        @property
        def args_default(self):
            "Returns the list of default arguments."
            return []

        @property
        def args_default_modified(self):
            "Returns the list of modified arguments."
            return []

        @property
        def args_mandatory(self):
            "Returns the list of mandatory arguments."
            return self.oinf.input_names

        @property
        def args_optional(self):
            "Returns the list of optional arguments."
            return []

        @property
        def obj(self):
            "Returns the ONNX graph."
            return self.oinf.obj

        def run(self, *args, **kwargs):
            "Calls run."
            return self.oinf.run(*args, **kwargs)

        def to_python(self, inputs, *args, **kwargs):
            "Calls to_python."
            res = self.oinf.to_python(*args, **kwargs)
            if len(res) != 1:
                raise NotImplementedError(  # pragma: no cover
                    "Not implemented if the code has multiple files.")
            keys = list(res)
            value = res[keys[0]]
            lines = value.split('\n')
            last = 0
            for i, line in enumerate(lines):
                if line.startswith('def '):
                    last = i - 1
                    break
            imports = '\n'.join(
                line for line in lines[:last] if 'import ' in line)
            lines.append('')
            lines.append(
                f"return OnnxPythonInference().run({', '.join(inputs)})")
            code = '\n'.join(lines[last:])
            return imports, code

        def need_context(self):
            "Needs context?"
            return False

        def enable_inplace_compute(self, index):
            "Not implemented."
            pass

    def __init__(self, onnx_node, desc, global_index):
        if desc is None:
            raise ValueError("desc should not be None.")  # pragma: no cover
        self.desc = desc
        self.onnx_node = onnx_node
        self._init(global_index)

    @property
    def name(self):
        "Returns the ONNX name."
        return "_".join(
            [self.desc['domain'], self.onnx_node.op_type]).replace(
                ".", "_").replace('__', '_').strip('_')

    def _init(self, global_index):
        """
        Prepares the node.
        """
        self.op_type = self.onnx_node.op_type
        self.order = -1
        self.variable_to_clean = []
        self.inputs = list(self.onnx_node.input)
        self.outputs = list(self.onnx_node.output)
        self.inplaces = []
        self.inputs_indices = [global_index(name) for name in self.inputs]
        self.outputs_indices = [global_index(name) for name in self.outputs]
        self._global_index = global_index

    def set_order(self, order):
        """
        Defines the order of execution.
        """
        self.order = order

    def add_variable_to_clean(self, name):
        """
        Adds a variable which can be cleaned after the node
        execution.
        """
        self.variable_to_clean.append(name)

    def __str__(self):
        "usual"
        return "Onnx-{}({}) -> {}{}".format(
            self.op_type, ", ".join(self.inputs), ", ".join(self.outputs),
            "    (name=%r)" % self.onnx_node.name
            if self.onnx_node.name else "")

    def __repr__(self):
        "usual"
        return self.__str__()

    def setup_runtime(self, runtime=None, variables=None, rt_class=None,
                      target_opset=None, dtype=None, domain=None,
                      ir_version=None, runtime_options=None,
                      build_inference_node_function=None,
                      existing_functions=None):
        """
        Loads runtime.

        :param runtime: runtime options
        :param variables: registered variables created by previous operators
        :param rt_class: runtime class used to compute
            prediction of subgraphs
        :param target_opset: use a specific target opset
        :param dtype: float computational type
        :param domain: node domain
        :param ir_version: if not None, changes the default value
            given by :epkg:`ONNX`
        :param runtime_options: runtime options
        :param build_inference_node_function: function creating an inference
            runtime from an ONNX graph
        :param existing_functions: existing function as a dictionary
            `{ (domain, name): fct }`

        .. versionchanged:: 0.9
            Parameters *build_inference_node_function* and *existing_functions*
            were added.
        """
        if self.desc is None:
            raise AttributeError(
                "desc should not be None.")  # pragma: no cover
        if rt_class is None:
            # path used when this operator is a function.
            self.function_ = OnnxInferenceNode.OnnxInferenceWrapper(runtime)
            self.ops_ = None
            return

        self.function_ = None
        self.preprocess_parameters(
            runtime, rt_class, ir_version=ir_version,
            target_opset=target_opset, existing_functions=existing_functions)
        options = {'provider': runtime} if runtime else {}
        if domain is not None:
            options['domain'] = domain
        if target_opset is not None:
            options['target_opset'] = target_opset
        if ir_version is not None:
            options['ir_version'] = ir_version
        if runtime_options is not None:
            options.update({
                k: v for k, v in runtime_options.items()
                if k not in {'log_severity_level'}})

        # existing functions?
        key = (self.onnx_node.domain, self.onnx_node.name)
        if existing_functions is not None and key in existing_functions:
            self.ops_ = existing_functions[key]
            return

        # regular node
        try:
            if runtime is not None and runtime.startswith('onnxruntime2'):
                self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                    options=options if options else None,
                                    variables=variables, dtype=dtype,
                                    runtime=runtime)
            elif runtime in ('python_compiled', 'python_compiled_debug'):
                options['provider'] = 'python'
                self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                    options=options if options else None,
                                    variables=variables, dtype=dtype,
                                    runtime=runtime)
            else:
                self.ops_ = load_op(self.onnx_node, desc=self.desc,
                                    options=options if options else None,
                                    variables=variables, dtype=dtype,
                                    runtime=runtime)
        except MissingOperatorError as e:
            try:
                onnx_schema = get_onnx_schema(
                    self.onnx_node.op_type, self.onnx_node.domain,
                    opset=target_opset)
            except SchemaError:
                fct_names = (
                    list(existing_functions.keys()) if existing_functions
                    else [])
                raise MissingOperatorError(
                    "Unable to find runtime for node (%r, %r), "
                    "available functions=%r." % (
                        self.onnx_node.domain, self.onnx_node.op_type,
                        fct_names)) from e
            if onnx_schema is None or not onnx_schema.has_function:
                raise e
            self.function_ = OnnxInferenceNode.OnnxInferenceWrapper(
                build_inference_node_function(onnx_schema.function_body))
            self.ops_ = None

    @staticmethod
    def _find_static_inputs(body):
        """
        Determines the loop inputs. It is any defined inputs
        by the subgraphs + any result used as a constant
        in the subgraphs.
        """
        inputs_set = set(i.name for i in body.input)
        for init in body.initializer:
            inputs_set.add(init.name)
        for node in body.node:
            for i in node.output:
                inputs_set.add(i)
        add_inputs = []
        for node in body.node:
            for i in node.input:
                if i not in inputs_set:
                    # no graph input or output node matches
                    # it must be a constant from the below graph
                    add_inputs.append(i)
                    inputs_set.add(i)
            for att in node.attribute:
                if (att.type == onnx_proto.AttributeProto.GRAPH and  # pylint: disable=E1101
                        hasattr(att, 'g') and att.g is not None):
                    inside = OnnxInferenceNode._find_static_inputs(att.g)
                    for i in inside:
                        if i not in inputs_set:
                            add_inputs.append(i)
                            inputs_set.add(i)
        # If there is no node, we add the outputs as well.
        if len(body.node) == 0:
            for o in body.output:
                i = o.name
                if i not in inputs_set:
                    add_inputs.append(i)
                    inputs_set.add(i)
        return add_inputs

    @staticmethod
    def _find_local_inputs(graph):
        """
        Determines the local inputs. It is any defined input
        used by the subgraph and defined in the parent graph.
        """
        if not isinstance(graph, GraphProto):
            raise TypeError(
                f"Unexpected type {type(graph)!r}.")
        local = set()
        known = set()
        for init in graph.initializer:
            known.add(init.name)
        for init in graph.input:
            known.add(init.name)
        for node in graph.node:
            for o in node.output:
                known.add(o)
            for i in node.input:
                if i not in known:
                    local.add(i)
        return list(local)

    def get_local_inputs(self):
        """
        Returns any local input used by this node in a subgraph
        defined as an attribute and not declared as an input of this subgraph.
        """
        req = set()
        for att in self.onnx_node.attribute:
            if hasattr(att, 'g') and att.g is not None:
                req |= set(self._find_local_inputs(att.g))
        return req

    def preprocess_parameters(self, runtime, rt_class, ir_version=None,
                              target_opset=None, existing_functions=None):
        """
        Preprocesses the parameters, loads *GraphProto*
        (equivalent to :epkg:`ONNX` graph with less metadata).

        :param runtime: runtime options
        :param rt_class: runtime class used to compute
            prediction of subgraphs
        :param ir_version: if not None, overwrites the default value
        :param target_opset: use a specific target opset
        :param existing_functions: existing functions
        """
        if 'atts' not in self.desc:
            return  # pragma: no cover
        inside_loop = self.onnx_node.op_type in {'Loop'}
        for _, v in self.desc['atts'].items():
            if 'value' not in v:
                continue  # pragma: no cover
            value = v['value']
            if isinstance(value, onnx_proto.GraphProto):
                static_inputs = OnnxInferenceNode._find_static_inputs(value)
                if len(value.node) > 0:
                    try:
                        sess = rt_class(value, runtime=runtime,
                                        ir_version=ir_version,
                                        target_opset=target_opset,
                                        inside_loop=inside_loop,
                                        static_inputs=static_inputs,
                                        existing_functions=existing_functions)
                    except RuntimeError as e:  # pragma: no cover
                        raise RuntimeError(
                            "Unable to instantiate a node of type %r and name %r."
                            "" % (self.onnx_node.op_type, self.onnx_node.name)) from e
                else:
                    # outputs already exists, usually branch then of else for If node
                    sess = rt_class(value, runtime=runtime,
                                    ir_version=ir_version,
                                    target_opset=target_opset,
                                    inside_loop=inside_loop,
                                    static_inputs=static_inputs,
                                    existing_functions=existing_functions)
                v['value_rt'] = sess

    def _build_context(self, values, input_list):
        context = {}
        # input_list does not need to be sorted but when
        # an input is not found, the returned error is always
        # related to the same input.
        for n in sorted(input_list):
            try:
                v = values[self._global_index(n)]
            except IndexError as e:
                raise IndexError(  # pragma: no cover
                    f"Unable to find an index for result {n!r} in onnx object.") from e
            if v is None:
                raise ValueError(  # pragma: no cover
                    f"Input {n!r} is None.")
            context[n] = v
        return context

    def run(self, values, attributes=None, verbose=0, fLOG=None):
        """
        Runs the node.
        The function updates values with outputs.

        :param values: list of existing values
        :param attributes: attributes known at function level
        :param verbose: verbosity
        :param fLOG: logging function
        """
        # This code takes time if the graph contains many nodes.
        # Maybe a C++ container would help in that case (to skip GIL).
        if self.inputs_indices is None:
            args = list(values[k] for k in self.inputs)
        else:
            args = list(values[k] for k in self.inputs_indices)

        if self.ops_ is None:
            # Then a function.
            if 'atts' in self.desc:
                # attributes of a function
                if attributes is None:
                    attributes = {}
                else:
                    attributes = attributes.copy()
                attributes.update(self.desc['atts'])

            feeds = {}
            for name, val in zip(self.function_.obj.input, args):
                if val is None:
                    raise ValueError(  # pragma: no cover
                        f"Input name {name!r} is None.")
                feeds[name] = val

            if verbose == 0 or fLOG is None:
                outputs = self.function_.run(feeds, attributes=attributes)
            else:
                if verbose > 0:
                    fLOG('-- >%s[%s](%s)  -- len(feeds)=%d' %
                         (self.function_.obj.name, self.function_.obj.domain,
                          ", ".join(self.function_.obj.input), len(feeds)))
                outputs = self.function_.run(
                    feeds, attributes=attributes, verbose=verbose, fLOG=fLOG)
                if verbose > 0:
                    fLOG('-- <%s[%s][%s]' %
                         (self.function_.obj.name, self.function_.obj.domain,
                          ", ".join(self.function_.obj.output)))

            res = [outputs[k] for k in self.function_.obj.output]
        else:
            # Or an operator.
            try:
                if self.ops_.need_context():
                    context = self._build_context(values,
                                                  self.ops_.additional_inputs)
                    res = self.ops_.run(*args, context=context,
                                        attributes=attributes,
                                        verbose=verbose, fLOG=fLOG)
                else:
                    res = self.ops_.run(
                        *args, attributes=attributes, verbose=verbose, fLOG=fLOG)
            except (ValueError, TypeError) as e:
                raise RuntimeError(  # pragma: no cover
                    "Unable to run operator %r, inputs=%r."
                    "" % (type(self.ops_), self.inputs)) from e
            except OverflowError as e:
                raise RuntimeError(  # pragma: no cover
                    "Unable to run operator %r, inputs=%r."
                    "" % (type(self.ops_), self.inputs)) from e

            if not isinstance(res, tuple):
                raise RuntimeError(  # pragma: no cover
                    f"Results of operator {type(self.ops_)!r} should be a tuple.")

        if len(self.outputs) != len(res):
            raise RuntimeError(  # pragma: no cover
                "Mismatch number of outputs got {} for names {}.\n{}".format(
                    len(res), list(sorted(self.outputs)),
                    pprint.pformat(self.desc)))

        # This code takes times if the graph contains many nodes.
        # Maybe a C++ container would help in that case (to skip GIL).
        if self.outputs_indices is None:
            for name, value in zip(self.outputs, res):
                values[name] = value
        else:
            for i, r in enumerate(res):
                values[self.outputs_indices[i]] = r

    def switch_initializers_dtype(self, dtype_in=numpy.float32,
                                  dtype_out=numpy.float64):
        """
        Switches all initializers to ``numpy.float64``.
        This only works if the runtime is ``'python'``.

        @param      dtype_in    previous type
        @param      dtype_out   next type
        @return                 done operations
        """
        done = []
        for k, v in self.desc['atts'].items():
            if 'value_rt' not in v:
                continue
            if isinstance(v['value_rt'], numpy.ndarray):
                if v['value_rt'].dtype == dtype_in:
                    v['value_rt'] = v['value_rt'].astype(dtype_out)
                    done.append(("+", "desc", k, v['value_rt']))
                else:
                    done.append(("-", "desc", k, v['value_rt']))
        if hasattr(self, 'ops_') and self.ops_ is not None:
            res = self.ops_.switch_initializers_dtype(dtype_in, dtype_out)
            for r in res:
                done.append(("ops_", ) + r)
        return done

    def enable_inplace_compute(self, name):
        """
        Let the node know that one input can be overwritten.

        @param      name        input name
        """
        self.inplaces.append(name)
        (self.ops_ or self.function_).enable_inplace_compute(
            self.inputs.index(name))

    @property
    def inputs_args(self):
        """
        Returns the list of arguments as well as
        the list of parameters with the default values
        (close to the signature).
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        sigs = []
        ops_or_function = self.function_ if self.ops_ is None else self.ops_
        mand = ops_or_function.args_mandatory
        if mand is None:
            mand = self.python_inputs
        sigs.extend(mand)
        if len(ops_or_function.args_optional) > 0:
            sigs.extend(ops_or_function.args_optional)
            if sys.version_info[:2] >= (3, 8):
                sigs.append('/')
        sigs.extend(ops_or_function.args_default)
        return sigs

    @property
    def python_inputs(self):
        """
        Returns the python arguments.
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        if hasattr(self.ops_, 'python_inputs'):
            return self.ops_.python_inputs
        return self.inputs

    @property
    def modified_args(self):
        """
        Returns the list of modified parameters.
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        if self.ops_ is None:
            return self.function_.args_default_modified
        return self.ops_.args_default_modified

    def to_python(self, inputs):
        """
        Returns a python code for this operator.

        @param      inputs      inputs name
        @return                 imports, python code, both as strings
        """
        if not hasattr(self, 'ops_'):
            raise AttributeError(
                "Attribute 'ops_' is missing.")  # pragma: no cover
        if self.ops_ is None:
            return self.function_.to_python(inputs)
        return self.ops_.to_python(inputs)



"""
@file
@brief Extensions to class @see cl OnnxInference.
"""
import os
import json
import re
from io import BytesIO
import pickle
import textwrap
from onnx import numpy_helper
from ..onnx_tools.onnx2py_helper import _var_as_dict, _type_to_string
from ..tools.graphs import onnx2bigraph
from ..plotting.text_plot import onnx_simple_text_plot


class OnnxInferenceExport:
    """
    Implements methods to export a instance of
    @see cl OnnxInference into :epkg:`json`, :epkg:`dot`,
    *text*, *python*.
    """

    def __init__(self, oinf):
        """
        @param      oinf    @see cl OnnxInference
        """
        self.oinf = oinf

    def to_dot(self, recursive=False, prefix='',  # pylint: disable=R0914
               add_rt_shapes=False, use_onnx=False,
               add_functions=True, **params):
        """
        Produces a :epkg:`DOT` language string for the graph.

        :param params: additional params to draw the graph
        :param recursive: also show subgraphs inside operator like
            @see cl Scan
        :param prefix: prefix for every node name
        :param add_rt_shapes: adds shapes infered from the python runtime
        :param use_onnx: use :epkg:`onnx` dot format instead of this one
        :param add_functions: add functions to the graph
        :return: string

        Default options for the graph are:

        ::

            options = {
                'orientation': 'portrait',
                'ranksep': '0.25',
                'nodesep': '0.05',
                'width': '0.5',
                'height': '0.1',
                'size': '7',
            }

        One example:

        .. exref::
            :title: Convert ONNX into DOT

            An example on how to convert an :epkg:`ONNX`
            graph into :epkg:`DOT`.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from mlprodict.npy.xop import loadop
                from mlprodict.onnxrt import OnnxInference

                OnnxAiOnnxMlLinearRegressor = loadop(
                    ('ai.onnx.ml', 'LinearRegressor'))

                pars = dict(coefficients=numpy.array([1., 2.]),
                            intercepts=numpy.array([1.]),
                            post_transform='NONE')
                onx = OnnxAiOnnxMlLinearRegressor(
                    'X', output_names=['Y'], **pars)
                model_def = onx.to_onnx(
                    {'X': pars['coefficients'].astype(numpy.float32)},
                    outputs={'Y': numpy.float32},
                    target_opset=12)
                oinf = OnnxInference(model_def)
                print(oinf.to_dot())

            See an example of representation in notebook
            :ref:`onnxvisualizationrst`.
        """
        clean_label_reg1 = re.compile("\\\\x\\{[0-9A-F]{1,6}\\}")
        clean_label_reg2 = re.compile("\\\\p\\{[0-9P]{1,6}\\}")

        def dot_name(text):
            return text.replace("/", "_").replace(
                ":", "__").replace(".", "_")

        def dot_label(text):
            for reg in [clean_label_reg1, clean_label_reg2]:
                fall = reg.findall(text)
                for f in fall:
                    text = text.replace(f, "_")  # pragma: no cover
            return text

        options = {
            'orientation': 'portrait',
            'ranksep': '0.25',
            'nodesep': '0.05',
            'width': '0.5',
            'height': '0.1',
            'size': '7',
        }
        options.update({k: v for k, v in params.items() if v is not None})

        if use_onnx:
            from onnx.tools.net_drawer import GetPydotGraph, GetOpNodeProducer

            pydot_graph = GetPydotGraph(
                self.oinf.obj.graph, name=self.oinf.obj.graph.name,
                rankdir=params.get('rankdir', "TB"),
                node_producer=GetOpNodeProducer(
                    "docstring", fillcolor="orange", style="filled",
                    shape="box"))
            return pydot_graph.to_string()

        inter_vars = {}
        exp = ["digraph{"]
        for opt in {'orientation', 'pad', 'nodesep', 'ranksep', 'size'}:
            if opt in options:
                exp.append(f"  {opt}={options[opt]};")
        fontsize = 10

        shapes = {}
        if add_rt_shapes:
            if not hasattr(self.oinf, 'shapes_'):
                raise RuntimeError(  # pragma: no cover
                    "No information on shapes, check the runtime '{}'."
                    "".format(self.oinf.runtime))
            for name, shape in self.oinf.shapes_.items():
                va = str(shape.shape)
                shapes[name] = va
                if name in self.oinf.inplaces_:
                    shapes[name] += "\\ninplace"

        # inputs
        exp.append("")
        graph = (
            self.oinf.obj.graph if hasattr(self.oinf.obj, 'graph')
            else self.oinf.obj)
        for obj in graph.input:
            if isinstance(obj, str):
                exp.append(
                    '  {2}{0} [shape=box color=red label="{0}" fontsize={1}];'
                    ''.format(obj, fontsize, prefix))
                inter_vars[obj] = obj
            else:
                dobj = _var_as_dict(obj)
                sh = shapes.get(dobj['name'], '')
                if sh:
                    sh = f"\\nshape={sh}"
                exp.append(
                    '  {3}{0} [shape=box color=red label="{0}\\n{1}{4}" fontsize={2}];'
                    ''.format(
                        dot_name(dobj['name']), _type_to_string(dobj['type']),
                        fontsize, prefix, dot_label(sh)))
                inter_vars[obj.name] = obj

        # outputs
        exp.append("")
        for obj in graph.output:
            if isinstance(obj, str):
                exp.append(
                    '  {2}{0} [shape=box color=green label="{0}" fontsize={1}];'.format(
                        obj, fontsize, prefix))
                inter_vars[obj] = obj
            else:
                dobj = _var_as_dict(obj)
                sh = shapes.get(dobj['name'], '')
                if sh:
                    sh = f"\\nshape={sh}"
                exp.append(
                    '  {3}{0} [shape=box color=green label="{0}\\n{1}{4}" fontsize={2}];'
                    ''.format(
                        dot_name(dobj['name']), _type_to_string(dobj['type']),
                        fontsize, prefix, dot_label(sh)))
                inter_vars[obj.name] = obj

        # initializer
        exp.append("")
        if hasattr(self.oinf.obj, 'graph'):
            inits = (
                list(self.oinf.obj.graph.initializer) +
                list(self.oinf.obj.graph.sparse_initializer))
            for obj in inits:
                dobj = _var_as_dict(obj)
                val = dobj['value']
                flat = val.flatten()
                if flat.shape[0] < 9:
                    st = str(val)
                else:
                    st = str(val)
                    if len(st) > 50:
                        st = st[:50] + '...'
                st = st.replace('\n', '\\n')
                kind = ""
                exp.append(
                    '  {6}{0} [shape=box label="{0}\\n{4}{1}({2})\\n{3}" fontsize={5}];'
                    ''.format(
                        dot_name(dobj['name']), dobj['value'].dtype,
                        dobj['value'].shape, dot_label(st), kind, fontsize, prefix))
                inter_vars[obj.name] = obj

        # nodes
        fill_names = {}
        if hasattr(self.oinf.obj, 'graph'):
            static_inputs = [n.name for n in self.oinf.obj.graph.input]
            static_inputs.extend(
                n.name for n in self.oinf.obj.graph.initializer)
            static_inputs.extend(
                n.name for n in self.oinf.obj.graph.sparse_initializer)
            nodes = self.oinf.obj.graph.node
        else:
            static_inputs = list(self.oinf.obj.input)
            nodes = self.oinf.obj.node
        for node in nodes:
            exp.append("")
            for out in node.output:
                if len(out) > 0 and out not in inter_vars:
                    inter_vars[out] = out
                    sh = shapes.get(out, '')
                    if sh:
                        sh = f"\\nshape={sh}"
                    exp.append(
                        '  {2}{0} [shape=box label="{0}{3}" fontsize={1}];'.format(
                            dot_name(out), fontsize, dot_name(prefix),
                            dot_label(sh)))
                static_inputs.append(out)

            dobj = _var_as_dict(node)
            if dobj['name'].strip() == '':  # pragma: no cover
                name = node.op_type
                iname = 1
                while name in fill_names:
                    name = "%s%d" % (name, iname)
                    iname += 1
                dobj['name'] = name
                node.name = name
                fill_names[name] = node

            atts = []
            if 'atts' in dobj:
                for k, v in sorted(dobj['atts'].items()):
                    val = None
                    if 'value' in v:
                        val = str(v['value']).replace(
                            "\n", "\\n").replace('"', "'")
                        sl = max(30 - len(k), 10)
                        if len(val) > sl:
                            val = val[:sl] + "..."
                    if val is not None:
                        atts.append(f'{k}={val}')
            satts = "" if len(atts) == 0 else ("\\n" + "\\n".join(atts))

            connects = []
            if recursive and node.op_type in {'Scan', 'Loop', 'If'}:
                fields = (['then_branch', 'else_branch']
                          if node.op_type == 'If' else ['body'])
                for field in fields:
                    if field not in dobj['atts']:
                        continue  # pragma: no cover

                    # creates the subgraph
                    body = dobj['atts'][field]['value']
                    oinf = self.oinf.__class__(
                        body, runtime=self.oinf.runtime,
                        skip_run=self.oinf.skip_run,
                        static_inputs=static_inputs)
                    subprefix = prefix + "B_"
                    subdot = oinf.to_dot(recursive=recursive, prefix=subprefix,
                                         add_rt_shapes=add_rt_shapes)
                    lines = subdot.split("\n")
                    start = 0
                    for i, line in enumerate(lines):
                        if '[' in line:
                            start = i
                            break
                    subgraph = "\n".join(lines[start:])

                    # connecting the subgraph
                    cluster = f"cluster_{node.op_type}{id(node)}_{id(field)}"
                    exp.append(f"  subgraph {cluster} {{")
                    exp.append('    label="{0}\\n({1}){2}";'.format(
                        dobj['op_type'], dot_name(dobj['name']), satts))
                    exp.append(f'    fontsize={fontsize};')
                    exp.append('    color=black;')
                    exp.append(
                        '\n'.join(map(lambda s: '  ' + s, subgraph.split('\n'))))

                    node0 = body.node[0]
                    connects.append((
                        f"{dot_name(subprefix)}{dot_name(node0.name)}",
                        cluster))

                    for inp1, inp2 in zip(node.input, body.input):
                        exp.append(
                            "  {0}{1} -> {2}{3};".format(
                                dot_name(prefix), dot_name(inp1),
                                dot_name(subprefix), dot_name(inp2.name)))
                    for out1, out2 in zip(body.output, node.output):
                        if len(out2) == 0:
                            # Empty output, it cannot be used.
                            continue
                        exp.append(
                            "  {0}{1} -> {2}{3};".format(
                                dot_name(subprefix), dot_name(out1.name),
                                dot_name(prefix), dot_name(out2)))
            else:
                exp.append('  {4}{1} [shape=box style="filled,rounded" color=orange '
                           'label="{0}\\n({1}){2}" fontsize={3}];'.format(
                               dobj['op_type'], dot_name(
                                   dobj['name']), satts, fontsize,
                               dot_name(prefix)))

            if connects is not None and len(connects) > 0:
                for name, cluster in connects:
                    exp.append(
                        "  {0}{1} -> {2} [lhead={3}];".format(
                            dot_name(prefix), dot_name(node.name),
                            name, cluster))

            for inp in node.input:
                exp.append(
                    "  {0}{1} -> {0}{2};".format(
                        dot_name(prefix), dot_name(inp), dot_name(node.name)))
            for out in node.output:
                if len(out) == 0:
                    # Empty output, it cannot be used.
                    continue
                exp.append(
                    "  {0}{1} -> {0}{2};".format(
                        dot_name(prefix), dot_name(node.name), dot_name(out)))

        if add_functions and len(self.oinf.functions_) > 0:
            for i, (k, v) in enumerate(self.oinf.functions_.items()):
                dot = v.to_dot(recursive=recursive, prefix=prefix + v.obj.name,
                               add_rt_shapes=add_rt_shapes,
                               use_onnx=use_onnx, add_functions=False,
                               **params)
                spl = dot.split('\n')[1:]
                exp.append('')
                exp.append('  subgraph cluster_%d {' % i)
                exp.append(f'    label="{v.obj.name}";')
                exp.append('    color=blue;')
                #exp.append('    style=filled;')
                exp.extend(('  ' + line) for line in spl)

        exp.append('}')
        return "\n".join(exp)

    def to_json(self, indent=2):
        """
        Converts an :epkg:`ONNX` model into :epkg:`JSON`.

        @param      indent      indentation
        @return                 string

        .. exref::
            :title: Convert ONNX into JSON

            An example on how to convert an :epkg:`ONNX`
            graph into :epkg:`JSON`.

            .. runpython::
                :showcode:
                :warningout: DeprecationWarning

                import numpy
                from mlprodict.npy.xop import loadop
                from mlprodict.onnxrt import OnnxInference

                OnnxAiOnnxMlLinearRegressor = loadop(
                    ('ai.onnx.ml', 'LinearRegressor'))

                pars = dict(coefficients=numpy.array([1., 2.]),
                            intercepts=numpy.array([1.]),
                            post_transform='NONE')
                onx = OnnxAiOnnxMlLinearRegressor(
                    'X', output_names=['Y'], **pars)
                model_def = onx.to_onnx(
                    {'X': pars['coefficients'].astype(numpy.float32)},
                    outputs={'Y': numpy.float32},
                    target_opset=12)
                oinf = OnnxInference(model_def)
                print(oinf.to_json())
        """

        def _to_json(obj):
            s = str(obj)
            rows = ['{']
            leave = None
            for line in s.split('\n'):
                if line.endswith("{"):
                    rows.append('"%s": {' % line.strip('{ '))
                elif ':' in line:
                    spl = line.strip().split(':')
                    if len(spl) != 2:
                        raise RuntimeError(  # pragma: no cover
                            f"Unable to interpret line '{line}'.")

                    if spl[0].strip() in ('type', ):
                        st = spl[1].strip()
                        if st in {'INT', 'INTS', 'FLOAT', 'FLOATS',
                                  'STRING', 'STRINGS', 'TENSOR'}:
                            spl[1] = f'"{st}"'

                    if spl[0] in ('floats', 'ints'):
                        if leave:
                            rows.append(f"{spl[1]},")
                        else:
                            rows.append(f'"{spl[0]}": [{spl[1].strip()},')
                            leave = spl[0]
                    elif leave:
                        rows[-1] = rows[-1].strip(',')
                        rows.append('],')
                        rows.append(f'"{spl[0].strip()}": {spl[1].strip()},')
                        leave = None
                    else:
                        rows.append(f'"{spl[0].strip()}": {spl[1].strip()},')
                elif line.strip() == "}":
                    rows[-1] = rows[-1].rstrip(",")
                    rows.append(line + ",")
                elif line:
                    raise RuntimeError(  # pragma: no cover
                        f"Unable to interpret line '{line}'.")
            rows[-1] = rows[-1].rstrip(',')
            rows.append("}")
            js = "\n".join(rows)

            try:
                content = json.loads(js)
            except json.decoder.JSONDecodeError as e:  # pragma: no cover
                js2 = "\n".join("%04d %s" % (i + 1, line)
                                for i, line in enumerate(js.split("\n")))
                raise RuntimeError(
                    f"Unable to parse JSON\n{js2}") from e
            return content

        # meta data
        final_obj = {}
        for k in {'ir_version', 'producer_name', 'producer_version',
                  'domain', 'model_version', 'doc_string'}:
            if hasattr(self.oinf.obj, k):
                final_obj[k] = getattr(self.oinf.obj, k)

        # inputs
        inputs = []
        for obj in self.oinf.obj.graph.input:
            st = _to_json(obj)
            inputs.append(st)
        final_obj['inputs'] = inputs

        # outputs
        outputs = []
        for obj in self.oinf.obj.graph.output:
            st = _to_json(obj)
            outputs.append(st)
        final_obj['outputs'] = outputs

        # init
        inits = {}
        for obj in self.oinf.obj.graph.initializer:
            value = numpy_helper.to_array(obj).tolist()
            inits[obj.name] = value
        final_obj['initializers'] = inits

        # nodes
        nodes = []
        for obj in self.oinf.obj.graph.node:
            node = dict(name=obj.name, op_type=obj.op_type, domain=obj.domain,
                        inputs=[str(_) for _ in obj.input],
                        outputs=[str(_) for _ in obj.output],
                        attributes={})
            for att in obj.attribute:
                st = _to_json(att)
                node['attributes'][st['name']] = st
                del st['name']
            nodes.append(node)
        final_obj['nodes'] = nodes

        return json.dumps(final_obj, indent=indent)

    def to_python(self, prefix="onnx_pyrt_", dest=None, inline=True):
        """
        Converts the ONNX runtime into independant python code.
        The function creates multiple files starting with
        *prefix* and saved to folder *dest*.

        @param  prefix      file prefix
        @param  dest        destination folder
        @param  inline      constant matrices are put in the python file itself
                            as byte arrays
        @return             file dictionary

        The function does not work if the chosen runtime
        is not *python*.

        .. runpython::
            :showcode:
            :warningout: DeprecationWarning

            import numpy
            from mlprodict.npy.xop import loadop
            from mlprodict.onnxrt import OnnxInference

            OnnxAdd = loadop('Add')

            idi = numpy.identity(2).astype(numpy.float32)
            onx = OnnxAdd('X', idi, output_names=['Y'],
                        op_version=12)
            model_def = onx.to_onnx({'X': idi},
                                    target_opset=12)
            X = numpy.array([[1, 2], [3, 4]], dtype=numpy.float32)
            oinf = OnnxInference(model_def, runtime='python')
            res = oinf.to_python()
            print(res['onnx_pyrt_main.py'])
        """
        if not isinstance(prefix, str):
            raise TypeError(  # pragma: no cover
                f"prefix must be a string not {type(prefix)!r}.")

        def clean_args(args):
            new_args = []
            for v in args:
                # remove python keywords
                if v.startswith('min='):
                    av = 'min_=' + v[4:]
                elif v.startswith('max='):
                    av = 'max_=' + v[4:]
                else:
                    av = v
                new_args.append(av)
            return new_args

        if self.oinf.runtime != 'python':
            raise ValueError(
                f"The runtime must be 'python' not '{self.oinf.runtime}'.")

        # metadata
        obj = {}
        for k in {'ir_version', 'producer_name', 'producer_version',
                  'domain', 'model_version', 'doc_string'}:
            if hasattr(self.oinf.obj, k):
                obj[k] = getattr(self.oinf.obj, k)
        code_begin = ["# coding: utf-8",
                      "'''",
                      "Python code equivalent to an ONNX graph.",
                      "It was was generated by module *mlprodict*.",
                      "'''"]
        code_imports = ["from io import BytesIO",
                        "import pickle",
                        "from numpy import array, float32, ndarray"]
        code_lines = ["class OnnxPythonInference:", "",
                      "    def __init__(self):",
                      "        self._load_inits()", "",
                      "    @property",
                      "    def metadata(self):",
                      f"        return {obj!r}", ""]

        # inputs
        if hasattr(self.oinf.obj, 'graph'):
            inputs = [obj.name for obj in self.oinf.obj.graph.input]
            outputs = [obj.name for obj in self.oinf.obj.graph.output]
        else:
            inputs = list(self.oinf.obj.input)
            outputs = list(self.oinf.obj.output)

        code_lines.extend([
            "    @property", "    def inputs(self):",
            f"        return {inputs!r}",
            ""
        ])

        # outputs
        code_lines.extend([
            "    @property", "    def outputs(self):",
            f"        return {outputs!r}",
            ""
        ])

        # init
        code_lines.extend(["    def _load_inits(self):",
                           "        self._inits = {}"])
        file_data = {}
        if hasattr(self.oinf.obj, 'graph'):
            for obj in self.oinf.obj.graph.initializer:
                value = numpy_helper.to_array(obj)
                bt = BytesIO()
                pickle.dump(value, bt)
                name = f'{prefix}{obj.name}.pkl'
                if inline:
                    code_lines.extend([
                        f"        iocst = {bt.getvalue()!r}",
                        f"        self._inits['{obj.name}'] = pickle.loads(iocst)"
                    ])
                else:
                    file_data[name] = bt.getvalue()
                    code_lines.append(
                        f"        self._inits['{obj.name}'] = pickle.loads('{name}')")
            code_lines.append('')

        # inputs, outputs
        inputs = self.oinf.input_names

        # nodes
        code_lines.extend([f"    def run(self, {', '.join(inputs)}):"])
        ops = {}
        if hasattr(self.oinf.obj, 'graph'):
            code_lines.append('        # constant')
            for obj in self.oinf.obj.graph.initializer:
                code_lines.append(
                    "        {0} = self._inits['{0}']".format(obj.name))
            code_lines.append('')
        code_lines.append('        # graph code')
        for node in self.oinf.sequence_:
            fct = 'pyrt_' + node.name
            if fct not in ops:
                ops[fct] = node
            args = []
            args.extend(node.inputs)
            margs = node.modified_args
            if margs is not None:
                args.extend(clean_args(margs))
            code_lines.append("        {0} = {1}({2})".format(
                ', '.join(node.outputs), fct, ', '.join(args)))
        code_lines.append('')
        code_lines.append('        # return')
        code_lines.append(f"        return {', '.join(outputs)}")
        code_lines.append('')

        # operator code
        code_nodes = []
        for name, op in ops.items():
            inputs_args = clean_args(op.inputs_args)

            code_nodes.append(f"def {name}({', '.join(inputs_args)}):")
            imps, code = op.to_python(op.python_inputs)
            if imps is not None:
                if not isinstance(imps, list):
                    imps = [imps]
                code_imports.extend(imps)
            code_nodes.append(textwrap.indent(code, '    '))
            code_nodes.extend(['', ''])

        # end
        code_imports = list(sorted(set(code_imports)))
        code_imports.extend(['', ''])
        file_data[prefix + 'main.py'] = "\n".join(
            code_begin + code_imports + code_nodes + code_lines)

        # saves as files
        if dest is not None:
            for k, v in file_data.items():
                ext = os.path.splitext(k)[-1]
                kf = os.path.join(dest, k)
                if ext == '.py':
                    with open(kf, "w", encoding="utf-8") as f:
                        f.write(v)
                elif ext == '.pkl':  # pragma: no cover
                    with open(kf, "wb") as f:
                        f.write(v)
                else:
                    raise NotImplementedError(  # pragma: no cover
                        f"Unknown extension for file '{k}'.")
        return file_data

    def to_text(self, recursive=False, grid=5, distance=5, kind='bi'):
        """
        It calls function @see fn onnx2bigraph to return
        the ONNX graph as text.

        :param recursive: dig into subgraphs too
        :param grid: align text to this grid
        :param distance: distance to the text
        :param kind: see below
        :return: text

        Possible values for format:
        * `'bi'`: use @see fn onnx2bigraph
        * `'seq'`: use @see fn onnx_simple_text_plot
        """
        if kind == 'bi':
            bigraph = onnx2bigraph(self.oinf.obj, recursive=recursive)
            graph = bigraph.display_structure(grid=grid, distance=distance)
            return graph.to_text()
        if kind == 'seq':
            return onnx_simple_text_plot(self.oinf.obj)
        raise ValueError(  # pragma: no cover
            f"Unexpected value for format={format!r}.")

    def to_onnx_code(self):
        """
        Exports the ONNX graph into an :epkg:`onnx` code
        which replicates it.

        :return: string
        """
        # Lazy import as it is not a common use.
        from ..onnx_tools.onnx_export import export2onnx
        return export2onnx(self.oinf.obj)



# pylint: disable=C0302
"""
@file
@brief Helpers for :epkg:`onnxruntime`.
"""


def get_ort_device(device):
    """
    Converts device into :epkg:`C_OrtDevice`.

    :param device: any type
    :return: :epkg:`C_OrtDevice`

    Example:

    ::

        get_ort_device('cpu')
        get_ort_device('gpu')
        get_ort_device('cuda')
        get_ort_device('cuda:0')
    """
    from onnxruntime.capi._pybind_state import (  # pylint: disable=E0611,W0611
        OrtDevice as C_OrtDevice)  # delayed
    if isinstance(device, C_OrtDevice):
        return device
    if isinstance(device, str):
        if device == 'cpu':
            return C_OrtDevice(
                C_OrtDevice.cpu(), C_OrtDevice.default_memory(), 0)
        if device in {'gpu', 'cuda:0', 'cuda', 'gpu:0'}:
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), 0)
        if device.startswith('gpu:'):
            idx = int(device[4:])
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), idx)
        if device.startswith('cuda:'):
            idx = int(device[5:])
            return C_OrtDevice(
                C_OrtDevice.cuda(), C_OrtDevice.default_memory(), idx)
        raise ValueError(  # pragma: no cover
            f"Unable to interpret string {device!r} as a device.")
    raise TypeError(  # pragma: no cover
        f"Unable to interpret type {type(device)!r}, ({device!r}) as de device.")


def device_to_providers(device):
    """
    Returns the corresponding providers for a specific device.

    :param device: :epkg:`C_OrtDevice`
    :return: providers
    """
    if isinstance(device, str):
        device = get_ort_device(device)
    if device.device_type() == device.cpu():
        return ['CPUExecutionProvider']
    if device.device_type() == device.cuda():
        return ['CUDAExecutionProvider', 'CPUExecutionProvider']
    raise ValueError(  # pragma: no cover
        f"Unexpected device {device!r}.")



# pylint: disable=W0611,R0401
"""
@file
@brief Shorcuts to plotting functions.
"""

from .text_plot import (
    onnx_text_plot, onnx_text_plot_tree, onnx_simple_text_plot,
    onnx_text_plot_io)
from .plotting_benchmark import plot_benchmark_metrics
from .plotting_validate_graph import plot_validate_benchmark
from .plotting_onnx import plot_onnx



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Helpers for operators Conv, ConvTranspose.
"""
import numpy
from .op_conv_helper_ import (  # pylint: disable=E0611
    im2col_1d_inplace_float,
    tch_im2col_2d_float, tch_col2im_2d_float,
    new_array as _new_array)


def im2col_nn(res):
    """
    Functions @see fn nn_im2col_2d and @see fn im2col returns the
    same results but with different shapes. This function
    converts a result from @see fn nn_im2col_2d into the same
    shape as a return from @see fn nn_im2col_2d.
    """
    if len(res.shape) % 2 != 0:
        raise ValueError(  # pragma: no cover
            "Number of dimensions should be even.")
    m = len(res.shape) // 2
    data = numpy.prod(res.shape[:m])
    ker = numpy.prod(res.shape[m:])
    resh = res.reshape((data, ker))
    tr = numpy.transpose(resh, [1, 0])
    return tr[numpy.newaxis, ...]


def new_array(shape, dtype=numpy.float32):
    """
    Creates a new empty array.

    :param shape: shape
    :param dtype: dtype
    :return: new array
    """
    if dtype == numpy.float32:
        dtype = numpy.dtype('float32')
    return _new_array(list(shape), dtype)


def nn_im2col_2d(data, kernel_shape, dilations, padding, fill_value=0):
    """
    C++ implementation for `im2col` or :func:`torch.nn.Unfold`.

    :param data: image (float), 2 dimensions.
    :param kernel_shape: kernel shape
    :param dilations: dilations
    :param padding: padding
    :param fill_value: fill value
    :return: result
    """
    strides = (1, 1)
    ext_shape = (
        (data.shape[0] + 2 * padding[0] - dilations[0] * (
            kernel_shape[0] - 1) - 1) // strides[0] + 1,
        (data.shape[1] + 2 * padding[1] - dilations[1] * (
            kernel_shape[1] - 1) - 1) // strides[1] + 1)
    kernel_size = kernel_shape[0] * kernel_shape[1]
    shape = (kernel_size, ext_shape[0] * ext_shape[1])
    result = numpy.empty(shape, dtype=data.dtype)
    if data.dtype == numpy.float32:
        tch_im2col_2d_float(result, data,
                            numpy.array(kernel_shape, dtype=numpy.int64),
                            numpy.array(dilations, dtype=numpy.int64),
                            numpy.array(padding, dtype=numpy.int64),
                            fill_value)
    else:
        raise NotImplementedError(  # pragma: no cover
            f"Unexpected dtype {data.dtype!r} for data.")
    return result


def nn_col2im_2d(data, output_shape, kernel_shape, dilations, padding):
    """
    C++ implementation for `col2im` or :func:`torch.nn.Fold`.

    :param data: image (float), 2 dimensions.
    :param output_shape: output size
    :param kernel_shape: kernel shape
    :param dilations: dilations
    :param padding: padding
    :return: result
    """
    result = numpy.empty(output_shape, dtype=data.dtype)
    if data.dtype == numpy.float32:
        tch_col2im_2d_float(result, data,
                            numpy.array(output_shape, dtype=numpy.int64),
                            numpy.array(kernel_shape, dtype=numpy.int64),
                            numpy.array(dilations, dtype=numpy.int64),
                            numpy.array(padding, dtype=numpy.int64))
    else:
        raise NotImplementedError(  # pragma: no cover
            f"Unexpected dtype {data.dtype!r} for data.")
    return result


def _get_indices(i, shape):
    res = numpy.empty((len(shape), ), dtype=numpy.int64)
    k = len(shape) - 1
    while k > 0:
        m = i % shape[k]
        res[k] = m
        i -= m
        i /= shape[k]
        k -= 1
    res[0] = i
    return res


def _is_out(ind, shape):
    for i, s in zip(ind, shape):
        if i < 0:
            return True
        if i >= s:
            return True
    return False


def im2col_naive_implementation(data, kernel_shape, fill_value=0):
    """
    Naive implementation for `im2col` or
    :func:`torch.nn.Unfold` (but with `padding=1`).

    :param image: image (float)
    :param kernel_shape: kernel shape
    :param fill_value: fill value
    :return: result
    """
    if not isinstance(kernel_shape, tuple):
        raise TypeError(
            f"Unexpected type {type(kernel_shape)!r} for kernel_shape.")
    if len(data.shape) != len(kernel_shape):
        raise ValueError(
            f"Shape mismatch {data.shape!r} and {kernel_shape!r}.")
    output_shape = data.shape + kernel_shape
    res = numpy.empty(output_shape, dtype=data.dtype)
    middle = numpy.array([-m / 2 for m in kernel_shape], dtype=numpy.int64)
    kernel_size = numpy.prod(kernel_shape)
    data_size = numpy.prod(data.shape)
    for i in range(data_size):
        for j in range(kernel_size):
            i_data = _get_indices(i, data.shape)
            i_kernel = _get_indices(j, kernel_shape)
            ind = i_data + i_kernel + middle
            t_data = tuple(i_data)
            t_kernel = tuple(i_kernel)
            i_out = t_data + t_kernel
            res[i_out] = fill_value if _is_out(
                ind, data.shape) else data[tuple(ind)]
    return res


def im2col_recursive(data, kernel_shape, fill_value=0, fall_back_dim=2):
    """
    Recursive implementation, falls back to
    @see fn im2col_naive_implementation for dimension `<= fall_back_dim`.
    The function is equivalent to
    :func:`torch.nn.Unfold` (but with `padding=1` on all dimensions).

    :param image: image (float)
    :param kernel_shape: kernel shape
    :param fill_value: fill value
    :param fall_back_dim: below that threshold,
        switches to @see fn im2col_naive_implementation.
    :return: result
    """
    if len(data.shape) <= fall_back_dim:
        return im2col_naive_implementation(data, kernel_shape, fill_value)

    perm = numpy.arange(len(data.shape) * 2).tolist()
    del perm[1:2]
    perm.insert(len(data.shape), 1)

    res = []
    N0 = data.shape[0]
    k0 = kernel_shape[0]
    mini_kernel = kernel_shape[1:]
    mini_shape = data.shape[1:] + mini_kernel
    for i in range(N0):
        for k in range(k0):
            ii = k - k0 // 2 + i
            if ii < 0 or ii >= N0:
                cc = numpy.full(mini_shape, dtype=data.dtype,
                                fill_value=fill_value)
            else:
                # many computation are already done, results should be cached.
                cc = im2col_recursive(data[ii], mini_kernel, fill_value)
            cc2 = cc[numpy.newaxis, ...]
            res.append(cc2)

    final = numpy.vstack(res)
    new_shape = (N0, k0) + cc.shape
    resh = final.reshape(new_shape)
    return numpy.transpose(resh, tuple(perm))


def im2col(data, kernel_shape=None, fill_value=0):
    """
    Returns the result of `im2col` on a image `NHCW` where N is 1.
    The function is equivalent to
    :func:`torch.nn.Unfold` (but with `padding=1` on all dimensions).

    :param image: image (float)
    :param kernel_shape: kernel shape
    :param fill_value: fill value
    :return: result

    This function is equivalent to function
    :func:`torch.nn.Unfold` with `padding=kernel_shape / 2`
    followed by a reshape and a transpose.

    ::

        import numpy
        from numpy.testing import assert_almost_equal
        import torch

        data = (numpy.arange(20).astype(numpy.float64) + 10).reshape((4, 5))
        expected = im2col_recursive(data, (3, 3), fill_value=0)
        unfold = torch.nn.Unfold(kernel_size=(3, 3), padding=1)
        input = torch.from_numpy(data.reshape((1, 1) + data.shape))
        output = unfold(input)
        mat = output.numpy()
        tr = numpy.transpose(mat, [0, 2, 1])
        resh = tr.reshape(expected.shape)
        assert_almost_equal(expected, resh)
    """
    if len(data.shape) == 1:
        if kernel_shape is None:
            kernel_shape = (3, )
        elif len(kernel_shape) != 1:
            raise ValueError(
                f"Unexpected kernel_shape {kernel_shape!r}, should be 1d.")
        if data.dtype == numpy.float32:
            result = numpy.empty(
                (data.shape[0], kernel_shape[0]), dtype=data.dtype)
            im2col_1d_inplace_float(
                result, data,
                kernel_shape if isinstance(kernel_shape, numpy.ndarray)
                else numpy.array(kernel_shape, dtype=numpy.int64),
                numpy.float32(fill_value))
            return result
    return im2col_naive_implementation(data, kernel_shape, fill_value)


def get_im2col_indices(x_shape, field_height, field_width, padding=1, stride=1):
    """
    Source `im2col.py <https://github.com/huyouare/CS231n/
    blob/master/assignment2/cs231n/im2col.py>`_.
    """
    # First figure out what the size of the output should be
    _, C, H, W = x_shape
    if (H + 2 * padding - field_height) % stride != 0:
        raise RuntimeError(
            "Unexpected value: %d != %d." % (
                H + 2 * padding - field_height, stride))
    if (W + 2 * padding - field_height) % stride != 0:
        raise RuntimeError(
            "Unexpected value: %d != %d." % (
                W + 2 * padding - field_height, stride))
    out_height = (H + 2 * padding - field_height) // stride + 1
    out_width = (W + 2 * padding - field_width) // stride + 1

    i0 = numpy.repeat(numpy.arange(field_height), field_width)
    i0 = numpy.tile(i0, C)
    i1 = stride * numpy.repeat(numpy.arange(out_height), out_width)
    j0 = numpy.tile(numpy.arange(field_width), field_height * C)
    j1 = stride * numpy.tile(numpy.arange(out_width), out_height)
    i = i0.reshape(-1, 1) + i1.reshape(1, -1)
    j = j0.reshape(-1, 1) + j1.reshape(1, -1)

    k = numpy.repeat(numpy.arange(C), field_height *
                     field_width).reshape(-1, 1)

    return (k, i, j)


def im2col_indices(x, field_height, field_width, padding=0, stride=1):
    """
    Source `im2col.py <https://github.com/huyouare/CS231n/
    blob/master/assignment2/cs231n/im2col.py>`_.
    """
    if padding > 0:
        p = padding
        x_padded = numpy.pad(
            x, ((0, 0), (0, 0), (p, p), (p, p)), mode='constant')
    else:
        x_padded = x
    k, i, j = get_im2col_indices(
        x.shape, field_height, field_width, padding, stride)
    cols = x_padded[:, k, i, j]
    C = x.shape[1]
    cols = cols.transpose(1, 2, 0).reshape(field_height * field_width * C, -1)
    return cols


def col2im_indices(cols, x_shape, field_height=3, field_width=3, padding=0,
                   stride=1):
    """
    Source `im2col.py <https://github.com/huyouare/CS231n/
    blob/master/assignment2/cs231n/im2col.py>`_.
    """
    N, C, H, W = x_shape
    H_padded, W_padded = H + 2 * padding, W + 2 * padding
    x_padded = numpy.zeros((N, C, H_padded, W_padded), dtype=cols.dtype)
    k, i, j = get_im2col_indices(x_shape, field_height, field_width, padding,
                                 stride)
    cols_reshaped = cols.reshape(C * field_height * field_width, -1, N)
    cols_reshaped = cols_reshaped.transpose(2, 0, 1)
    numpy.add.at(x_padded, (slice(None), k, i, j), cols_reshaped)
    if padding == 0:
        return x_padded
    return x_padded[:, :, padding:-padding, padding:-padding]



# -*- encoding: utf-8 -*-
# pylint: disable=E0203,E1101,C0111
"""
@file
@brief Runtime operator.
"""
import numpy
from ._op import OpRun


class PRelu(OpRun):

    def __init__(self, onnx_node, desc=None, **options):
        OpRun.__init__(self, onnx_node, desc=desc,
                       **options)

    def _run(self, x, slope, attributes=None, verbose=0, fLOG=None):  # pylint: disable=W0221
        return (numpy.where(x > 0, x, x * slope), )

    def to_python(self, inputs):
        return ('import numpy',
                "return numpy.where({0} > 0, {0}, {0} * {1})".format(
                    inputs[0], inputs[1]))



"""
@file
@brief Shortcut to *cli*.
"""
from .convert_validate import convert_validate
from .einsum import einsum_test
from .onnx_code import onnx_code
from .optimize import onnx_optim
from .validate import validate_runtime, latency
